"""Chemprop integration tests."""
from flask import url_for
from io import BytesIO
import json
import os
from tempfile import TemporaryDirectory
from typing import List
import unittest
from unittest import TestCase
from unittest.mock import patch

import numpy as np
import pandas as pd
from parameterized import parameterized

from chemprop.constants import TEST_SCORES_FILE_NAME
from chemprop.hyperparameter_optimization import chemprop_hyperopt
from chemprop.interpret import chemprop_interpret
from chemprop.sklearn_predict import sklearn_predict
from chemprop.sklearn_train import sklearn_train
from chemprop.train import chemprop_train, chemprop_predict, evaluate_predictions, chemprop_fingerprint
from chemprop.web.wsgi import build_app
from chemprop.spectra_utils import normalize_spectra, load_phase_mask
from chemprop.features import load_features


TEST_DATA_DIR = 'tests/data'
SEED = 0
EPOCHS = 10
NUM_FOLDS = 3
NUM_ITER = 2
SIZE = 10
DEPTH = 2
DELTA = 0.025


class ChempropTests(TestCase):
    @staticmethod
    def create_raw_train_args(dataset_type: str,
                              metric: str,
                              save_dir: str,
                              model_type: str = 'chemprop',
                              flags: List[str] = None) -> List[str]:
        """Creates a list of raw command line arguments for training."""
        return [
            'train',  # Note: not actually used, just a placeholder
            '--data_path', os.path.join(TEST_DATA_DIR, f'{dataset_type}.csv'), # Note: adding another --data_path argument will overwrite this one
            '--dataset_type', dataset_type,
            '--epochs', str(EPOCHS),
            '--num_folds', str(NUM_FOLDS),
            '--seed', str(SEED),
            '--metric', metric,
            '--save_dir', save_dir,
            '--hidden_size', str(SIZE),
            '--ffn_hidden_size', str(SIZE),
            '--depth', str(DEPTH),
            '--quiet',
            '--empty_cache'
        ] + (['--model_type', model_type] if model_type != 'chemprop' else []) + (flags if flags is not None else [])

    @staticmethod
    def create_raw_predict_args(dataset_type: str,
                                preds_path: str,
                                checkpoint_dir: str,
                                flags: List[str] = None) -> List[str]:
        """Creates a list of raw command line arguments for predicting."""
        return [
            'predict',  # Note: not actually used, just a placeholder
            '--test_path', os.path.join(TEST_DATA_DIR, f'{dataset_type}_test_smiles.csv'),
            '--preds_path', preds_path,
            '--checkpoint_dir', checkpoint_dir
        ] + (flags if flags is not None else [])

    @staticmethod
    def create_raw_hyperopt_args(dataset_type: str,
                                 config_save_path: str,
                                 save_dir: str,
                                 flags: List[str] = None) -> List[str]:
        """Creates a list of raw command line arguments for hyperparameter optimization."""
        return [
            'hyperopt',  # Note: not actually used, just a placeholder
            '--data_path', os.path.join(TEST_DATA_DIR, f'{dataset_type}.csv'),
            '--dataset_type', dataset_type,
            '--epochs', str(EPOCHS),
            '--num_iter', str(NUM_ITER),
            '--config_save_path', config_save_path,
            '--save_dir', save_dir,
            '--quiet',
            '--empty_cache'
        ] + (flags if flags is not None else [])

    @staticmethod
    def create_raw_interpret_args(dataset_type: str,
                                  checkpoint_dir: str,
                                  flags: List[str] = None) -> List[str]:
        """Creates a list of raw command line arguments for interpretation."""
        return [
            'interpret',  # Note: not actually used, just a placeholder
            '--data_path', os.path.join(TEST_DATA_DIR, f'{dataset_type}_test_smiles.csv'),
            '--checkpoint_dir', checkpoint_dir
        ] + (flags if flags is not None else [])

    def train(self,
              dataset_type: str,
              metric: str,
              save_dir: str,
              model_type: str = 'chemprop',
              flags: List[str] = None):
        # Set up command line arguments
        raw_args = self.create_raw_train_args(
            dataset_type=dataset_type,
            metric=metric,
            save_dir=save_dir,
            model_type=model_type,
            flags=flags
        )

        # Train
        with patch('sys.argv', raw_args):
            command_line = ' '.join(raw_args[1:])

            if model_type == 'chemprop':
                print(f'python train.py {command_line}')
                chemprop_train()
            else:
                print(f'python sklearn_train.py {command_line}')
                sklearn_train()

    def predict(self,
                dataset_type: str,
                preds_path: str,
                save_dir: str,
                model_type: str = 'chemprop',
                flags: List[str] = None):
        # Set up command line arguments
        raw_args = self.create_raw_predict_args(
            dataset_type=dataset_type,
            preds_path=preds_path,
            checkpoint_dir=save_dir,
            flags=flags
        )

        # Predict
        with patch('sys.argv', raw_args):
            command_line = ' '.join(raw_args[1:])

            if model_type == 'chemprop':
                print(f'python predict.py {command_line}')
                chemprop_predict()
            else:
                print(f'python sklearn_predict.py {command_line}')
                sklearn_predict()

    def hyperopt(self,
                 dataset_type: str,
                 config_save_path: str,
                 save_dir: str,
                 flags: List[str] = None):
        # Set up command line arguments
        raw_args = self.create_raw_hyperopt_args(
            dataset_type=dataset_type,
            config_save_path=config_save_path,
            save_dir=save_dir,
            flags=flags
        )

        # Hyperopt
        with patch('sys.argv', raw_args):
            command_line = ' '.join(raw_args[1:])
            print(f'python hyperparameter_optimization.py {command_line}')
            chemprop_hyperopt()

    def interpret(self,
                  dataset_type: str,
                  checkpoint_dir: str,
                  flags: List[str] = None):
        # Set up command line arguments
        raw_args = self.create_raw_interpret_args(
            dataset_type=dataset_type,
            checkpoint_dir=checkpoint_dir,
            flags=flags
        )

        # Interpret
        with patch('sys.argv', raw_args):
            command_line = ' '.join(raw_args[1:])
            print(f'python interpret.py {command_line}')
            chemprop_interpret()

    def fingerprint(self,
                    dataset_type: str,
                    checkpoint_dir: str,
                    fingerprint_path: str,
                    fingerprint_flags: List[str]):
        # Set up command line arguments
        raw_args = self.create_raw_predict_args(
            dataset_type=dataset_type,
            preds_path=fingerprint_path,
            checkpoint_dir=checkpoint_dir,
            flags=fingerprint_flags
        )

        # Fingerprint
        with patch('sys.argv', raw_args):
            command_line = ' '.join(raw_args[1:])
            print(f'python fingerprint.py {command_line}')
            chemprop_fingerprint()

    @parameterized.expand([
        (
                'sklearn_random_forest',
                'random_forest',
                'rmse',
                1.582733
        ),
        (
                'sklearn_svm',
                'svm',
                'rmse',
                1.698927
        ),
        (
                'chemprop',
                'chemprop',
                'rmse',
                2.338310289,
        ),
        (
                'chemprop_scaffold_split',
                'chemprop',
                'rmse',
                2.2983865,
                ['--split_type', 'scaffold_balanced']
        ),
        (
                'chemprop_morgan_features_generator',
                'chemprop',
                'rmse',
                2.0438637,
                ['--features_generator', 'morgan']
        ),
        (
                'sklearn_random_forest_rdkit_features_path',
                'random_forest',
                'rmse',
                0.691494,
                ['--features_path', os.path.join(TEST_DATA_DIR, 'regression.npz'), '--no_features_scaling']
        ),
        (
                'sklearn_svm_rdkit_features_path',
                'svm',
                'rmse',
                1.022634,
                ['--features_path', os.path.join(TEST_DATA_DIR, 'regression.npz'), '--no_features_scaling']
        ),
        (
                'chemprop_rdkit_features_path',
                'chemprop',
                'rmse',
                2.14015989,
                ['--features_path', os.path.join(TEST_DATA_DIR, 'regression.npz'), '--no_features_scaling', '--save_smiles_splits']
        ),
        (
                'chemprop_features_generator_features_path',
                'chemprop',
                'rmse',
                1.59283050,
                ['--features_generator', 'morgan', '--features_path', os.path.join(TEST_DATA_DIR, 'regression.npz'), '--no_features_scaling']
        ),
        (
                'chemprop_bounded_mse_loss',
                'chemprop',
                'bounded_mse',
                5.10242,
                [
                    '--loss_function', 'bounded_mse',
                    '--data_path', os.path.join(TEST_DATA_DIR, 'regression_inequality.csv')
                ]
        )
    ])
    def test_train_single_task_regression(self,
                                          name: str,
                                          model_type: str,
                                          metric: str,
                                          expected_score: float,
                                          train_flags: List[str] = None):
        with TemporaryDirectory() as save_dir:
            # Train
            self.train(
                dataset_type='regression',
                metric=metric,
                save_dir=save_dir,
                model_type=model_type,
                flags=train_flags
            )

            # Check results
            test_scores_data = pd.read_csv(os.path.join(save_dir, TEST_SCORES_FILE_NAME))
            test_scores = np.array(test_scores_data[f'Mean {metric}'])
            self.assertEqual(len(test_scores), 1)

            mean_score = np.mean(test_scores)
            self.assertAlmostEqual(mean_score, expected_score, delta=DELTA*expected_score)

    @parameterized.expand([
        (
                'chemprop',
                'chemprop',
                'auc',
                0.52783634,
                ['--class_balance', '--split_sizes', '0.4', '0.3', '0.3']
        ),
        (
                'chemprop_morgan_features_generator',
                'chemprop',
                'auc',
                0.519689086,
                ['--features_generator', 'morgan', '--class_balance', '--split_sizes', '0.4', '0.3', '0.3']
        ),
        (
                'chemprop_rdkit_features_path',
                'chemprop',
                'auc',
                0.466828424,
                ['--features_path', os.path.join(TEST_DATA_DIR, 'classification.npz'), '--no_features_scaling', '--class_balance', '--split_sizes', '0.4', '0.3', '0.3', '--save_smiles_splits']
        ),
        (
                'chemprop_features_generator_features_path',
                'chemprop',
                'auc',
                0.499183589,
                ['--features_generator', 'morgan', '--features_path', os.path.join(TEST_DATA_DIR, 'classification.npz'), '--no_features_scaling', '--class_balance', '--split_sizes', '0.4', '0.3', '0.3']
        ),
        (
                'chemprop_mcc_metric',
                'chemprop',
                'mcc',
                0.014589067,
                ['--metric', 'mcc', '--data_path', os.path.join(TEST_DATA_DIR, 'classification_common.csv'), '--class_balance']
        ),
        (
                'chemprop_f1_metric',
                'chemprop',
                'f1',
                0.190841899,
                ['--metric', 'f1', '--data_path', os.path.join(TEST_DATA_DIR, 'classification_common.csv'), '--class_balance']
        ),
        (
                'chemprop_mcc_loss',
                'chemprop',
                'auc',
                0.55505265,
                ['--loss_function', 'mcc', '--data_path', os.path.join(TEST_DATA_DIR, 'classification_common.csv'), '--class_balance']
        )
    ])
    def test_train_multi_task_classification(self,
                                             name: str,
                                             model_type: str,
                                             metric: str,
                                             expected_score: float,
                                             train_flags: List[str] = None):
        with TemporaryDirectory() as save_dir:
            # Train
            self.train(
                dataset_type='classification',
                metric=metric,
                save_dir=save_dir,
                model_type=model_type,
                flags=train_flags
            )

            # Check results
            test_scores_data = pd.read_csv(os.path.join(save_dir, TEST_SCORES_FILE_NAME))
            test_scores = np.array(test_scores_data[f'Mean {metric}'])
            mean_score = np.mean(np.array(test_scores))
            self.assertAlmostEqual(mean_score, expected_score, delta=DELTA*expected_score)

    @parameterized.expand([
        (
                'sklearn_random_forest',
                'random_forest',
                0.9455894
        ),
        (
                'sklearn_svm',
                'svm',
                1.0151356
        ),
        (
                'chemprop',
                'chemprop',
                2.4145471
        ),
        (
                'chemprop_morgan_features_generator',
                'chemprop',
                2.4703284,
                ['--features_generator', 'morgan'],
                ['--features_generator', 'morgan']
        ),
        (
                'sklearn_random_forest_rdkit_features_path',
                'random_forest',
                0.2954347,
                ['--features_path', os.path.join(TEST_DATA_DIR, 'regression.npz'), '--no_features_scaling'],
                ['--features_path', os.path.join(TEST_DATA_DIR, 'regression_test.npz'), '--no_features_scaling']
        ),
        (
                'sklearn_svm_rdkit_features_path',
                'svm',
                0.4112432,
                ['--features_path', os.path.join(TEST_DATA_DIR, 'regression.npz'), '--no_features_scaling'],
                ['--features_path', os.path.join(TEST_DATA_DIR, 'regression_test.npz'), '--no_features_scaling']
        ),
        (
                'chemprop_rdkit_features_path',
                'chemprop',
                1.51978455,
                ['--features_path', os.path.join(TEST_DATA_DIR, 'regression.npz'), '--no_features_scaling', '--save_smiles_splits'],
                ['--features_path', os.path.join(TEST_DATA_DIR, 'regression_test.npz'), '--no_features_scaling']
        ),
        (
                'chemprop_features_generator_features_path',
                'chemprop',
                0.59545263,
                ['--features_generator', 'morgan', '--features_path', os.path.join(TEST_DATA_DIR, 'regression.npz'), '--no_features_scaling'],
                ['--features_generator', 'morgan', '--features_path', os.path.join(TEST_DATA_DIR, 'regression_test.npz'), '--no_features_scaling']
        )
    ])
    def test_predict_single_task_regression(self,
                                            name: str,
                                            model_type: str,
                                            expected_score: float,
                                            train_flags: List[str] = None,
                                            predict_flags: List[str] = None):
        with TemporaryDirectory() as save_dir:
            # Train
            dataset_type = 'regression'
            self.train(
                dataset_type=dataset_type,
                metric='rmse',
                save_dir=save_dir,
                model_type=model_type,
                flags=train_flags
            )

            # Predict
            preds_path = os.path.join(save_dir, 'preds.csv')
            self.predict(
                dataset_type=dataset_type,
                preds_path=preds_path,
                save_dir=save_dir,
                model_type=model_type,
                flags=predict_flags
            )

            # Check results
            pred = pd.read_csv(preds_path)
            true = pd.read_csv(os.path.join(TEST_DATA_DIR, f'{dataset_type}_test_true.csv'))
            self.assertEqual(list(pred.keys()), list(true.keys()))
            self.assertEqual(list(pred['smiles']), list(true['smiles']))

            pred, true = pred.drop(columns=['smiles']), true.drop(columns=['smiles'])
            pred, true = pred.to_numpy(), true.to_numpy()
            mse = float(np.mean((pred - true) ** 2))
            self.assertAlmostEqual(mse, expected_score, delta=DELTA*expected_score)

    def test_predict_individual_ensemble(self):
        with TemporaryDirectory() as save_dir:
            # Train
            dataset_type = 'regression'
            self.train(
                dataset_type=dataset_type,
                metric='rmse',
                save_dir=save_dir,
            )

            # Predict
            preds_path = os.path.join(save_dir, 'preds.csv')
            self.predict(
                dataset_type=dataset_type,
                preds_path=preds_path,
                save_dir=save_dir,
                flags=['--individual_ensemble_predictions']
            )

            pred = pd.read_csv(preds_path)
            columns = list(pred.columns)
            expected_columns = ['smiles', 'logSolubility'] + [f'logSolubility_model_{idx}' for idx in range(NUM_FOLDS)]
            self.assertTrue(columns == expected_columns)


    @parameterized.expand([
        (
                'chemprop',
                'chemprop',
                0.238491358,
                ['--class_balance', '--split_sizes', '0.4', '0.3', '0.3'],
        ),
        (
                'chemprop_morgan_features_generator',
                'chemprop',
                0.254056869,
                ['--features_generator', 'morgan', '--class_balance', '--split_sizes', '0.4', '0.3', '0.3'],
                ['--features_generator', 'morgan']
        ),
        (
                'chemprop_rdkit_features_path',
                'chemprop',
                0.307159229,
                ['--features_path', os.path.join(TEST_DATA_DIR, 'classification.npz'), '--no_features_scaling', '--class_balance', '--split_sizes', '0.4', '0.3', '0.3', '--save_smiles_splits'],
                ['--features_path', os.path.join(TEST_DATA_DIR, 'classification_test.npz'), '--no_features_scaling']
        ),
        (
                'chemprop_features_generator_features_path',
                'chemprop',
                0.193924687,
                ['--features_generator', 'morgan', '--features_path', os.path.join(TEST_DATA_DIR, 'classification.npz'), '--no_features_scaling', '--class_balance', '--split_sizes', '0.4', '0.3', '0.3'],
                ['--features_generator', 'morgan', '--features_path', os.path.join(TEST_DATA_DIR, 'classification_test.npz'), '--no_features_scaling']
        )
    ])
    def test_predict_multi_task_classification(self,
                                               name: str,
                                               model_type: str,
                                               expected_score: float,
                                               train_flags: List[str] = None,
                                               predict_flags: List[str] = None):
        with TemporaryDirectory() as save_dir:
            # Train
            dataset_type = 'classification'
            self.train(
                dataset_type=dataset_type,
                metric='auc',
                save_dir=save_dir,
                model_type=model_type,
                flags=train_flags
            )

            # Predict
            preds_path = os.path.join(save_dir, 'preds.csv')
            self.predict(
                dataset_type=dataset_type,
                preds_path=preds_path,
                save_dir=save_dir,
                model_type=model_type,
                flags=predict_flags
            )

            # Check results
            pred = pd.read_csv(preds_path)
            true = pd.read_csv(os.path.join(TEST_DATA_DIR, f'{dataset_type}_test_true.csv'))
            self.assertEqual(list(pred.keys()), list(true.keys()))
            self.assertEqual(list(pred['smiles']), list(true['smiles']))

            pred, true = pred.drop(columns=['smiles']), true.drop(columns=['smiles'])
            pred, true = pred.to_numpy(), true.to_numpy()
            mse = float(np.nanmean((pred - true) ** 2))
            self.assertAlmostEqual(mse, expected_score, delta=DELTA*expected_score)

    def test_chemprop_hyperopt(self):
        with TemporaryDirectory() as save_dir:
            # Train
            config_save_path = os.path.join(save_dir, 'config.json')
            self.hyperopt(
                dataset_type='regression',
                config_save_path=config_save_path,
                save_dir=save_dir
            )

            # Check results
            with open(config_save_path) as f:
                config = json.load(f)

            parameters = {'depth': (2, 6), 'hidden_size': (300, 2400), 'ffn_hidden_size': (300, 2400), 'ffn_num_layers': (1, 3), 'dropout': (0.0, 0.4)}

            self.assertEqual(set(config.keys()), set(parameters.keys()))

            for parameter, (min_value, max_value) in parameters.items():
                self.assertTrue(min_value <= config[parameter] <= max_value)

    @parameterized.expand([
        (
                'chemprop',
        ),
        (
                'chemprop_morgan_features_generator',
                ['--features_generator', 'morgan'],
                ['--features_generator', 'morgan']
        ),
    ])
    def test_interpret_single_task_regression(self,
                                              name: str,
                                              train_flags: List[str] = None,
                                              interpret_flags: List[str] = None):
        with TemporaryDirectory() as save_dir:
            # Train
            dataset_type = 'regression'
            self.train(
                dataset_type=dataset_type,
                metric='rmse',
                save_dir=save_dir,
                flags=train_flags
            )

            # Interpret
            try:
                self.interpret(
                    dataset_type=dataset_type,
                    checkpoint_dir=save_dir,
                    flags=interpret_flags
                )
            except Exception as e:
                self.fail(f'Interpretation failed with error: {e}')

    def test_chemprop_web(self):
        with TemporaryDirectory() as root_dir:
            app = build_app(root_folder=root_dir, init_db=True)

            app.config['TESTING'] = True
            app.config['SERVER_NAME'] = 'localhost'

            data_path = 'regression.csv'
            test_path = 'regression_test_smiles.csv'
            dataset_name = 'regression_data'
            dataset_type = 'regression'
            checkpoint_name = 'regression_ckpt'
            ckpt_name = data_name = '1'
            epochs = 3
            ensemble_size = 1

            with open(os.path.join(TEST_DATA_DIR, data_path)) as f:
                train_data = BytesIO(f.read().encode('utf-8'))

            with open(os.path.join(TEST_DATA_DIR, test_path)) as f:
                test_smiles = f.read()

            with app.app_context():
                with app.test_client() as client:
                    response = client.get('/')
                    self.assertEqual(response.status_code, 200)

                    # Upload data
                    response = client.post(
                        url_for('upload_data', return_page='home'),
                        data={
                            'dataset': (train_data, data_path),
                            'datasetName': dataset_name
                        }
                    )
                    self.assertEqual(response.status_code, 302)

                    # Train
                    response = client.post(
                        url_for('train'),
                        data={
                            'dataName': data_name,
                            'epochs': epochs,
                            'ensembleSize': ensemble_size,
                            'checkpointName': checkpoint_name,
                            'datasetType': dataset_type,
                            'useProgressBar': False
                        }
                    )
                    self.assertEqual(response.status_code, 200)

                    # Predict
                    response = client.post(
                        url_for('predict'),
                        data={
                            'checkpointName': ckpt_name,
                            'textSmiles': test_smiles
                        }
                    )
                    self.assertEqual(response.status_code, 200)

    @parameterized.expand([
        (
            'spectra',
            'chemprop',
            0.09920149,
            [
                '--data_path', os.path.join(TEST_DATA_DIR, 'spectra.csv'),
                '--features_path', os.path.join(TEST_DATA_DIR, 'spectra_features.csv'),
                '--split_type', 'random_with_repeated_smiles', '--save_smiles_splits'
            ]
        ),
        (
            'spectra_excluded_targets',
            'chemprop',
            0.08912992,
            [
                '--data_path', os.path.join(TEST_DATA_DIR, 'spectra_exclusions.csv'),
                '--features_path', os.path.join(TEST_DATA_DIR, 'spectra_features.csv'),
                '--split_type', 'random_with_repeated_smiles', '--save_smiles_splits'
            ]
        ),
        (
            'spectra_phase_features',
            'chemprop',
            0.0747605825,
            [
                '--data_path', os.path.join(TEST_DATA_DIR, 'spectra_exclusions.csv'),
                '--phase_features_path', os.path.join(TEST_DATA_DIR, 'spectra_features.csv'),
                '--spectra_phase_mask_path', os.path.join(TEST_DATA_DIR, 'spectra_mask.csv'),
                '--split_type', 'random_with_repeated_smiles'
            ]
        ),
    ])
    def test_train_spectra(self,
                                          name: str,
                                          model_type: str,
                                          expected_score: float,
                                          train_flags: List[str] = None):
        with TemporaryDirectory() as save_dir:
            # Train
            metric = 'sid'
            self.train(
                dataset_type = 'spectra',
                metric = metric,
                save_dir = save_dir,
                model_type = model_type,
                flags = train_flags
            )

            # Check results
            test_scores_data = pd.read_csv(os.path.join(save_dir, TEST_SCORES_FILE_NAME))
            test_scores = np.array(test_scores_data[f'Mean {metric}'])
            self.assertEqual(len(test_scores), 1)

            mean_score = np.mean(test_scores)
            self.assertAlmostEqual(mean_score, expected_score, delta=DELTA*expected_score)

    @parameterized.expand([
        (
            'spectra',
            'chemprop',
            0.074686983,
            0,
            [
                '--data_path', os.path.join(TEST_DATA_DIR, 'spectra.csv'),
                '--features_path', os.path.join(TEST_DATA_DIR, 'spectra_features.csv'),
                '--split_type', 'random_with_repeated_smiles', '--save_smiles_splits'
            ],
            [
                '--features_path', os.path.join(TEST_DATA_DIR, 'spectra_features.csv'),
            ]
        ),
        (
            'spectra_phase_features',
            'chemprop',
            0.076007918,
            284,
            [
                '--data_path', os.path.join(TEST_DATA_DIR, 'spectra_exclusions.csv'),
                '--phase_features_path', os.path.join(TEST_DATA_DIR, 'spectra_features.csv'),
                '--spectra_phase_mask_path', os.path.join(TEST_DATA_DIR, 'spectra_mask.csv'),
                '--split_type', 'random_with_repeated_smiles'
            ],
            [
                '--phase_features_path', os.path.join(TEST_DATA_DIR, 'spectra_features.csv'),
            ]
        ),
    ])
    def test_predict_spectra(self,
                                            name: str,
                                            model_type: str,
                                            expected_score: float,
                                            expected_nans: int,
                                            train_flags: List[str] = None,
                                            predict_flags: List[str] = None):
        with TemporaryDirectory() as save_dir:
            # Train
            dataset_type = 'spectra'
            self.train(
                dataset_type=dataset_type,
                metric='sid',
                save_dir=save_dir,
                model_type=model_type,
                flags=train_flags
            )

            # Predict
            preds_path = os.path.join(save_dir, 'preds.csv')
            self.predict(
                dataset_type=dataset_type,
                preds_path=preds_path,
                save_dir=save_dir,
                model_type=model_type,
                flags=predict_flags
            )

            # Check results
            pred = pd.read_csv(preds_path)
            true = pd.read_csv(os.path.join(TEST_DATA_DIR, 'spectra.csv'))
            self.assertEqual(list(pred.keys()), list(true.keys()))
            self.assertEqual(list(pred['smiles']), list(true['smiles']))

            pred, true = pred.drop(columns=['smiles']), true.drop(columns=['smiles'])
            pred, true = pred.to_numpy(), true.to_numpy()
            phase_features = load_features(predict_flags[1])
            if '--spectra_phase_mask_path' in train_flags:
                mask = load_phase_mask(train_flags[5])
            else:
                mask = None
            true = normalize_spectra(true,phase_features,mask)
            sid = evaluate_predictions(preds=pred, targets=true, num_tasks=len(true[0]), metrics=['sid'], dataset_type='spectra')['sid'][0]
            self.assertAlmostEqual(sid, expected_score, delta=DELTA*expected_score)
            self.assertEqual(np.sum(np.isnan(pred)),expected_nans)

    @parameterized.expand([
        (
                'chemprop_reaction',
                'chemprop',
                2.3338595,
                ['--reaction', '--data_path', os.path.join(TEST_DATA_DIR, 'reaction_regression.csv')]
        ),
        (
                'chemprop_scaffold_split',
                'chemprop',
                2.095871,
                ['--reaction', '--data_path', os.path.join(TEST_DATA_DIR, 'reaction_regression.csv'),'--split_type', 'scaffold_balanced']
        ),
        (
                'chemprop_morgan_features_generator',
                'chemprop',
                3.122113679,
                ['--reaction', '--data_path', os.path.join(TEST_DATA_DIR, 'reaction_regression.csv'),'--features_generator', 'morgan']
        ),
        (
                'chemprop_reaction_explicit_h',
                'chemprop',
                2.34536046,
                ['--reaction', '--data_path', os.path.join(TEST_DATA_DIR, 'reaction_regression.csv'), '--explicit_h']
         )
    ])
    def test_train_single_task_regression_reaction(self,
                                          name: str,
                                          model_type: str,
                                          expected_score: float,
                                          train_flags: List[str] = None):
        with TemporaryDirectory() as save_dir:
            # Train
            metric = 'rmse'
            self.train(
                dataset_type = 'regression',
                metric = metric,
                save_dir = save_dir,
                model_type = model_type,
                flags = train_flags
            )

            # Check results
            test_scores_data = pd.read_csv(os.path.join(save_dir, TEST_SCORES_FILE_NAME))
            test_scores = np.array(test_scores_data[f'Mean {metric}'])
            self.assertEqual(len(test_scores), 1)

            mean_score = np.mean(test_scores)
            self.assertAlmostEqual(mean_score, expected_score, delta=DELTA*expected_score)

    @parameterized.expand([
        (
                'chemprop',
                'chemprop',
                'auc',
                0.6644781145,
                ['--number_of_molecules', '2', '--data_path', os.path.join(TEST_DATA_DIR, 'classification_multimolecule.csv')]
        )
    ])
    def test_single_task_multimolecule_classification(self,
                                          name: str,
                                          model_type: str,
                                          metric: str,
                                          expected_score: float,
                                          train_flags: List[str] = None):
        with TemporaryDirectory() as save_dir:
            # Train
            self.train(
                dataset_type='classification',
                metric=metric,
                save_dir=save_dir,
                model_type=model_type,
                flags=train_flags
            )

            # Check results
            test_scores_data = pd.read_csv(os.path.join(save_dir, TEST_SCORES_FILE_NAME))
            test_scores = np.array(test_scores_data[f'Mean {metric}'])

            mean_score = np.mean(test_scores)
            self.assertAlmostEqual(mean_score, expected_score, delta=DELTA * expected_score)

    @parameterized.expand([
        (
                'chemprop',
                'chemprop',
                31.669185768,
                ['--fingerprint_type', 'MPN'],
                ['--split_sizes', '0.4', '0.3', '0.3'],
        ),
        (
                'chemprop',
                'chemprop',
                26.903715076,
                ['--fingerprint_type', 'last_FFN'],
                ['--split_sizes', '0.4', '0.3', '0.3']
        )
    ])
    def test_single_task_fingerprint(self,
                                            name: str,
                                            model_type: str,
                                            expected_score: float,
                                            fingerprint_flags: List[str],
                                            train_flags: List[str] = None,
                                     ):
        with TemporaryDirectory() as save_dir:
            # Train
            dataset_type = 'classification'
            self.train(
                dataset_type=dataset_type,
                metric='auc',
                save_dir=save_dir,
                model_type=model_type,
                flags=train_flags
            )

            # Fingerprint
            fingerprint_path = os.path.join(save_dir, 'fingerprints.csv')
            self.fingerprint(
                dataset_type=dataset_type,
                checkpoint_dir=save_dir,
                fingerprint_path=fingerprint_path,
                fingerprint_flags=fingerprint_flags
            )

            fingerprints = pd.read_csv(fingerprint_path).drop(["smiles"], axis=1)
            self.assertAlmostEqual(np.sum(fingerprints.to_numpy()), expected_score, delta=DELTA*expected_score)

    @parameterized.expand([
        (
                'chemprop',
                'chemprop',
                ['--fingerprint_type', 'MPN'],
                True,
                ['--number_of_molecules', '2', '--data_path', os.path.join(TEST_DATA_DIR, 'classification_multimolecule.csv')]
        ),
        (
                'chemprop',
                'chemprop',
                ['--fingerprint_type', 'MPN'],
                False,
                ['--number_of_molecules', '2', '--data_path', os.path.join(TEST_DATA_DIR, 'classification_multimolecule.csv'), '--mpn_shared']
        ),
        (
                'chemprop',
                'chemprop',
                ['--fingerprint_type', 'last_FFN'],
                True,
                ['--number_of_molecules', '2', '--data_path', os.path.join(TEST_DATA_DIR, 'classification_multimolecule.csv')]
        )
    ])
    def test_multimolecule_fingerprint_with_single_input(self,
                                     name: str,
                                     model_type: str,
                                     fingerprint_flags: List[str],
                                     exception_thrown: bool,
                                     train_flags: List[str] = None
                                     ):
        with TemporaryDirectory() as save_dir:
            # Train
            dataset_type = 'classification'
            self.train(
                dataset_type=dataset_type,
                metric='auc',
                save_dir=save_dir,
                model_type=model_type,
                flags=train_flags
            )

            # Predict
            fingerprint_path = os.path.join(save_dir, 'fingerprints.csv')

            # Check to make sure that an exception is thrown for cases where the model isn't built with --mpn-shared and with a fingerprint
            # type of MPN
            if exception_thrown:
                with self.assertRaises(ValueError):
                    self.fingerprint(
                        dataset_type=dataset_type,
                        checkpoint_dir=save_dir,
                        fingerprint_path=fingerprint_path,
                        fingerprint_flags=fingerprint_flags
                    )
            else:
                self.fingerprint(
                    dataset_type=dataset_type,
                    checkpoint_dir=save_dir,
                    fingerprint_path=fingerprint_path,
                    fingerprint_flags=fingerprint_flags
                )
                fingerprints = pd.read_csv(fingerprint_path)
                test_input = pd.read_csv(os.path.join(TEST_DATA_DIR, f'{dataset_type}_test_smiles.csv'))

                self.assertEqual(list(fingerprints['smiles'].values), list(test_input['smiles'].values))

    @parameterized.expand([(
                'chemprop_reaction_solvent',
                'chemprop',
                3.0195756,
                ['--reaction_solvent', '--number_of_molecules', '2',
                 '--data_path', os.path.join(TEST_DATA_DIR, 'reaction_solvent_regression.csv')]
        ),
        (
                'chemprop_morgan_features_generator',
                'chemprop',
                3.34822937696,
                ['--reaction_solvent', '--number_of_molecules', '2',
                 '--data_path', os.path.join(TEST_DATA_DIR, 'reaction_solvent_regression.csv'),'--features_generator', 'morgan']
        ),
        (
                'chemprop_reaction_solvent_explicit_h',
                'chemprop',
                3.01241762,
                ['--reaction_solvent', '--number_of_molecules', '2',
                 '--data_path', os.path.join(TEST_DATA_DIR, 'reaction_solvent_regression.csv'), '--explicit_h']
         ),
        (
                'chemprop_reaction_solvent_explicit_h_adding_h',
                'chemprop',
                2.984292677,
                ['--reaction_solvent', '--number_of_molecules', '2',
                 '--data_path', os.path.join(TEST_DATA_DIR, 'reaction_solvent_regression.csv'), '--explicit_h', '--adding_h']
        ),
        (
                'chemprop_reaction_solvent_diff_mpn_size',
                'chemprop',
                2.734318,
                ['--reaction_solvent', '--number_of_molecules', '2',
                 '--data_path', os.path.join(TEST_DATA_DIR, 'reaction_solvent_regression.csv'), '--hidden_size', '500',
                 '--hidden_size_solvent', '250']
        )
    ])
    def test_train_single_task_regression_reaction_solvent(self,
                                          name: str,
                                          model_type: str,
                                          expected_score: float,
                                          train_flags: List[str] = None):
        with TemporaryDirectory() as save_dir:
            # Train
            metric = 'rmse'
            self.train(
                dataset_type = 'regression',
                metric = metric,
                save_dir = save_dir,
                model_type = model_type,
                flags = train_flags
            )

            # Check results
            test_scores_data = pd.read_csv(os.path.join(save_dir, TEST_SCORES_FILE_NAME))
            test_scores = np.array(test_scores_data[f'Mean {metric}'])
            self.assertEqual(len(test_scores), 1)

            mean_score = np.mean(test_scores)
            self.assertAlmostEqual(mean_score, expected_score, delta=DELTA*expected_score)

    @parameterized.expand([(
        9450.06996,
        'ensemble',
        None,
        'nll',
        [],
        [],
    ),
    (
        -2.0835368,
        'mve',
        None,
        'nll',
        ['--loss_function', 'mve'],
        [],
    ),
    (
        -2.12500003,
        'evidential_epistemic',
        None,
        'nll',
        ['--loss_function', 'evidential'],
        [],
    ),
    (
        -2.0854637,
        'evidential_aleatoric',
        None,
        'nll',
        ['--loss_function', 'evidential'],
        [],
    ),
    (
        -2.0130289,
        'evidential_total',
        None,
        'nll',
        ['--loss_function', 'evidential'],
        [],
    ),
    (
        0.7223842,
        'dropout',
        'zscaling',
        'ence',
        ['--num_folds', '1'],
        [],
    ),
    (
        -1.9783182,
        'ensemble',
        'zscaling',
        'nll',
        [],
        [],
    ),
    (
        6.94374243,
        'ensemble',
        'tscaling',
        'nll',
        ['--num_folds','5'],
        [],
    ),
    (
        0.480757667,
        'ensemble',
        'zelikman_interval',
        'ence',
        [],
        [],
    ),
    (
        -2.0896678,
        'mve',
        'mve_weighting',
        'nll',
        ['--loss_function', 'mve'],
        [],
    ),
    (
        0.486563126,
        'ensemble',
        None,
        'miscalibration_area',
        [],
        [],
    ),
    (
        99.40899,
        'ensemble',
        None,
        'ence',
        [],
        [],
    ),
    # (
    #     0.0239197,
    #     'ensemble',
    #     None,
    #     'spearman',
    #     [],
    #     [],
    # ),
    (
        0.903807,
        None,
        "conformal_regression",
        "conformal_coverage",
        [],
        ["--conformal_alpha", "0.1"]
    ),
    (
        0.703407,
        None,
        "conformal_regression",
        "conformal_coverage",
        [],
        ["--conformal_alpha", "0.3"]
    ),
    (
        0.902953,
        None,
        "conformal_quantile_regression",
        "conformal_coverage",
        ["--loss_function", "quantile_interval", "--quantile_loss_alpha", "0.1"],
        ["--conformal_alpha", "0.1"]
    ),
    (
        0.702532,
        None,
        "conformal_quantile_regression",
        "conformal_coverage",
        ["--loss_function", "quantile_interval", "--quantile_loss_alpha", "0.3"],
        ["--conformal_alpha", "0.3"]
    )
    ])
    def test_uncertainty_regression(
        self,
        expected_score: float,
        uncertainty_method: str,
        calibration_method: str,
        evaluation_methods: str,
        train_flags: List[str] = None,
        predict_flags: List[str] = None,
    ):
        with TemporaryDirectory() as save_dir:
            data_path = os.path.join(TEST_DATA_DIR, 'regression_multitask.csv')
            train_flags.extend(['--data_path', data_path])

            self.train(
                dataset_type='regression',
                metric='rmse',
                save_dir=save_dir,
                flags=train_flags,
            )

            test_path = os.path.join(TEST_DATA_DIR, 'regression_multitask_gaps.csv')
            eval_path = os.path.join(save_dir, 'eval_scores.csv')
            predict_flags.extend(['--evaluation_scores_path', eval_path, '--test_path', test_path])
            if uncertainty_method is not None:
                predict_flags.extend(['--uncertainty_method', uncertainty_method,])
            if calibration_method is not None:
                predict_flags.extend(['--calibration_method', calibration_method, '--calibration_path', test_path])
            if evaluation_methods is not None:
                predict_flags.extend(['--evaluation_methods', evaluation_methods])
            self.predict(
                dataset_type='regression',
                preds_path=os.path.join(save_dir, 'preds.csv'),
                save_dir=save_dir,
                flags=predict_flags,
            )
            evaluation_scores_data=pd.read_csv(eval_path)

            self.assertAlmostEqual(evaluation_scores_data['homo'][0], expected_score, delta=np.abs(expected_score * DELTA))

    @parameterized.expand([(
        0.66787529,
        'classification',
        'platt',
        'nll',
        ['--number_of_molecules', '2'],
        ['--number_of_molecules', '2'],
    ),
    (
        0.6411087455,
        'classification',
        'isotonic',
        'nll',
        ['--number_of_molecules', '2'],
        ['--number_of_molecules', '2'],
    ),
    (
        0.6254826255,
        'classification',
        'isotonic',
        'accuracy',
        ['--number_of_molecules', '2'],
        ['--number_of_molecules', '2'],
    ),
    ])
    def test_uncertainty_class(
        self,
        expected_score: float,
        uncertainty_method: str,
        calibration_method: str,
        evaluation_methods: str,
        train_flags: List[str] = None,
        predict_flags: List[str] = None,
    ):
        with TemporaryDirectory() as save_dir:
            test_path = os.path.join(TEST_DATA_DIR, 'classification_multimolecule.csv')
            train_flags.extend(['--data_path', test_path])
            self.train(
                dataset_type='classification',
                metric='binary_cross_entropy',
                save_dir=save_dir,
                flags=train_flags,
            )
            eval_path = os.path.join(save_dir, 'eval_scores.csv')
            predict_flags.extend(['--evaluation_scores_path', eval_path, '--test_path', test_path])
            if uncertainty_method is not None:
                predict_flags.extend(['--uncertainty_method', uncertainty_method,])
            if calibration_method is not None:
                predict_flags.extend(['--calibration_method', calibration_method, '--calibration_path', test_path])
            if evaluation_methods is not None:
                predict_flags.extend(['--evaluation_methods', evaluation_methods])
            self.predict(
                dataset_type='regression',
                preds_path=os.path.join(save_dir, 'preds.csv'),
                save_dir=save_dir,
                flags=predict_flags,
            )
            evaluation_scores_data=pd.read_csv(eval_path)
            self.assertAlmostEqual(evaluation_scores_data['synergy'][0], expected_score, delta=expected_score * DELTA)

    @parameterized.expand(
        [
            (
                0.900013,
                None,
                "conformal",
                "conformal_coverage",
                [],
                ["--conformal_alpha", "0.1"]),
            (
                0.900013,
                None, 
                "conformal_adaptive",
                "conformal_coverage",
                [],
                ["--conformal_alpha", "0.1"]
            ),
        ]
    )
    def test_conformal_multiclass(
        self,
        expected_score: float,
        uncertainty_method: str,
        calibration_method: str,
        evaluation_methods: str,
        train_flags: List[str] = None,
        predict_flags: List[str] = None,
    ):
        with TemporaryDirectory() as save_dir:
            data_path = os.path.join(TEST_DATA_DIR, "multiclass.csv")
            train_flags.extend(["--data_path", data_path, "--multiclass_num_classes", "11"])

            self.train(
                dataset_type="multiclass",
                metric="cross_entropy",
                save_dir=save_dir,
                flags=train_flags,
            )
            eval_path = os.path.join(save_dir, "eval_scores.csv")
            test_path = os.path.join(TEST_DATA_DIR, "multiclass.csv")
            preds_path = os.path.join(save_dir, "preds.csv")
            predict_flags.extend(["--evaluation_scores_path", eval_path, "--test_path", test_path])
            if uncertainty_method is not None:
                predict_flags.extend(["--uncertainty_method", uncertainty_method])
            if calibration_method is not None:
                predict_flags.extend(
                    ["--calibration_method", calibration_method, "--calibration_path", test_path]
                )
            if evaluation_methods is not None:
                predict_flags.extend(["--evaluation_methods", evaluation_methods])
            self.predict(
                dataset_type="multiclass",
                preds_path=preds_path,
                save_dir=save_dir,
                flags=predict_flags,
            )
            evaluation_scores = (
                pd.read_csv(eval_path).drop(columns=["evaluation_method"]).to_numpy()
            )
            mean_score = np.mean(evaluation_scores)
            self.assertAlmostEqual(mean_score, expected_score, delta=DELTA*expected_score)

    @parameterized.expand(
        [
            (
                0.994667,
                None,
                "conformal",
                "conformal_coverage",
                [],
                ["--conformal_alpha", "0.1"],
            ),
            (
                0.994667,
                None,
                "conformal",
                "conformal_coverage",
                [],
                ["--conformal_alpha", "0.3"],
            )
        ]
    )
    def test_conformal_multilabel(
        self,
        expected_score: float,
        uncertainty_method: str,
        calibration_method: str,
        evaluation_methods: str,
        train_flags: List[str] = None,
        predict_flags: List[str] = None,
    ):
        with TemporaryDirectory() as save_dir:
            data_path = os.path.join(TEST_DATA_DIR, "classification.csv")
            train_flags.extend(["--data_path", data_path])

            self.train(
                dataset_type="classification", metric="auc", save_dir=save_dir, flags=train_flags
            )
            eval_path = os.path.join(save_dir, "eval_scores.csv")
            test_path = os.path.join(TEST_DATA_DIR, "classification.csv")
            preds_path = os.path.join(save_dir, "preds.csv")
            predict_flags.extend(["--evaluation_scores_path", eval_path, "--test_path", test_path])
            if uncertainty_method is not None:
                predict_flags.extend(["--uncertainty_method", uncertainty_method])
            if calibration_method is not None:
                predict_flags.extend(
                    ["--calibration_method", calibration_method, "--calibration_path", test_path]
                )
            if evaluation_methods is not None:
                predict_flags.extend(["--evaluation_methods", evaluation_methods])
            self.predict(
                dataset_type="classification",
                preds_path=preds_path,
                save_dir=save_dir,
                flags=predict_flags,
            )
            evaluation_scores = (
                pd.read_csv(eval_path).drop(columns=["evaluation_method"]).to_numpy()
            )
            mean_score = np.mean(evaluation_scores)
            self.assertAlmostEqual(mean_score, expected_score, delta=DELTA*expected_score)

    @parameterized.expand([
        (
                'chemprop_atomic_bond_targets',
                'chemprop',
                8.710007,
                ['--data_path', os.path.join(TEST_DATA_DIR, 'atomic_bond_regression.csv'),
                 '--is_atom_bond_targets',
                 '--adding_h']
        ),
        (
                'chemprop_atomic_bond_targets_constraints',
                'chemprop',
                8.435722,
                ['--data_path', os.path.join(TEST_DATA_DIR, 'atomic_bond_regression.csv'),
                 '--constraints_path', os.path.join(TEST_DATA_DIR, 'atomic_bond_constraints.csv'),
                 '--is_atom_bond_targets',
                 '--adding_h']
        ),
        (
                'chemprop_atomic_bond_targets_no_shared_atom_bond_ffn',
                'chemprop',
                8.619740,
                ['--data_path', os.path.join(TEST_DATA_DIR, 'atomic_bond_regression.csv'),
                 '--is_atom_bond_targets',
                 '--adding_h',
                 '--no_shared_atom_bond_ffn']
        ),
        (
                'chemprop_atomic_bond_targets_constraints_no_adding_bond_types',
                'chemprop',
                8.780137,
                ['--data_path', os.path.join(TEST_DATA_DIR, 'atomic_bond_regression.csv'),
                 '--is_atom_bond_targets',
                 '--adding_h',
                 '--no_adding_bond_types']
        ),
        (
                'chemprop_atomic_bond_targets_weights_ffn_num_layers',
                'chemprop',
                8.765641,
                ['--data_path', os.path.join(TEST_DATA_DIR, 'atomic_bond_regression.csv'),
                 '--constraints_path', os.path.join(TEST_DATA_DIR, 'atomic_bond_constraints.csv'),
                 '--is_atom_bond_targets',
                 '--adding_h',
                 '--weights_ffn_num_layers', '3']
        ),
        (
                'chemprop_atomic_targets_atom_mapped',
                'chemprop',
                0.233874,
                ['--data_path', os.path.join(TEST_DATA_DIR, 'atomic_regression_atom_mapped.csv'),
                 '--is_atom_bond_targets',
                 '--keeping_atom_map']
        )
    ])
    def test_train_multi_task_regression_atomic_bond_targets(self,
                                                             name: str,
                                                             model_type: str,
                                                             expected_score: float,
                                                             train_flags: List[str] = None):
        with TemporaryDirectory() as save_dir:
            # Train
            metric = 'rmse'
            self.train(
                dataset_type = 'regression',
                metric = metric,
                save_dir = save_dir,
                model_type = model_type,
                flags = train_flags
            )

            # Check results
            test_scores_data = pd.read_csv(os.path.join(save_dir, TEST_SCORES_FILE_NAME))
            test_scores = test_scores_data[f'Mean {metric}']
            mean_score = test_scores.mean()
            self.assertAlmostEqual(mean_score, expected_score, delta=DELTA * expected_score)


if __name__ == '__main__':
    unittest.main()


from chemprop.args import HyperoptArgs
import os
import pickle
from typing import List, Dict
import csv
import json
import logging

from hyperopt import Trials, hp
import numpy as np

from chemprop.constants import HYPEROPT_SEED_FILE_NAME
from chemprop.utils import makedirs


def build_search_space(search_parameters: List[str], train_epochs: int = None) -> dict:
    """
    Builds the parameter space to be searched with hyperopt trials.

    :param search_parameters: A list of parameters to be included in the search space.
    :param train_epochs: The total number of epochs to be used in training.
    :return: A dictionary keyed by the parameter names of hyperopt search functions.
    """
    available_spaces = {
        "activation": hp.choice(
            "activation", options=["ReLU", "LeakyReLU", "PReLU", "tanh", "SELU", "ELU"]
        ),
        "aggregation": hp.choice("aggregation", options=["mean", "sum", "norm"]),
        "aggregation_norm": hp.quniform("aggregation_norm", low=1, high=200, q=1),
        "batch_size": hp.quniform("batch_size", low=5, high=200, q=5),
        "depth": hp.quniform("depth", low=2, high=6, q=1),
        "dropout": hp.quniform("dropout", low=0.0, high=0.4, q=0.05),
        "ffn_hidden_size": hp.quniform("ffn_hidden_size", low=300, high=2400, q=100),
        "ffn_num_layers": hp.quniform("ffn_num_layers", low=1, high=3, q=1),
        "final_lr_ratio": hp.loguniform("final_lr_ratio", low=np.log(1e-4), high=0.),
        "hidden_size": hp.quniform("hidden_size", low=300, high=2400, q=100),
        "init_lr_ratio": hp.loguniform("init_lr_ratio", low=np.log(1e-4), high=0.),
        "linked_hidden_size": hp.quniform("linked_hidden_size", low=300, high=2400, q=100),
        "max_lr": hp.loguniform("max_lr", low=np.log(1e-6), high=np.log(1e-2)),
        "warmup_epochs": hp.quniform("warmup_epochs", low=1, high=train_epochs // 2, q=1)
    }
    space = {}
    for key in search_parameters:
        space[key] = available_spaces[key]

    return space


def merge_trials(trials: Trials, new_trials_data: List[Dict]) -> Trials:
    """
    Merge a hyperopt trials object with the contents of another hyperopt trials object.

    :param trials: A hyperopt trials object containing trials data, organized into hierarchical dictionaries.
    :param trials_data: The contents of a hyperopt trials object, `Trials.trials`.
    :return: A hyperopt trials object, merged from the two inputs.
    """
    if len(trials.trials) > 0:
        max_tid = max([trial["tid"] for trial in trials.trials])
        trial_keys = set(trials.vals.keys())
        for trial in trials.trials:
            new_trial_keys = set(trial["misc"]["vals"].keys())
            if trial_keys != new_trial_keys:
                raise ValueError(
                    f"Hyperopt trials with different search spaces cannot be combined. \
                        Across the loaded previous trials, the parameters {trial_keys} \
                        were included in the search space of some trials. At least one \
                        trial includes only the parameters {new_trial_keys}."
                )
    else:
        trial_keys = None
        max_tid = 0

    for trial in new_trials_data:
        new_trial_keys = set(trial["misc"]["vals"].keys())
        if trial_keys is None:
            trial_keys = new_trial_keys
        elif new_trial_keys != trial_keys:
            raise ValueError(
                f"Hyperopt trials with different search spaces cannot be combined. \
                    A new trial searching for parameters {new_trial_keys} was merged \
                    with another trial for parameters {trial_keys}"
            )
        tid = (
            trial["tid"] + max_tid + 1
        )  # trial id needs to be unique among this list of ids.
        hyperopt_trial = Trials().new_trial_docs(
            tids=[None], specs=[None], results=[None], miscs=[None]
        )
        hyperopt_trial[0] = trial
        hyperopt_trial[0]["tid"] = tid
        hyperopt_trial[0]["misc"]["tid"] = tid
        for key in hyperopt_trial[0]["misc"]["idxs"].keys():
            hyperopt_trial[0]["misc"]["idxs"][key] = [tid]
        trials.insert_trial_docs(hyperopt_trial)
        trials.refresh()
    return trials


def load_trials(dir_path: str, previous_trials: Trials = None) -> Trials:
    """
    Load in trials from each pickle file in the hyperopt checkpoint directory.
    Checkpoints are newly loaded in at each iteration to allow for parallel entries
    into the checkpoint folder by independent hyperoptimization instances.

    :param dir_path: Path to the directory containing hyperopt checkpoint files.
    :param previous_trials: Any previously generated trials objects that the loaded trials will be merged with.
    :return: A trials object containing the merged trials from all checkpoint files.
    """

    # List out all the pickle files in the hyperopt checkpoint directory
    hyperopt_checkpoint_files = [
        os.path.join(dir_path, path) for path in os.listdir(dir_path) if ".pkl" in path
    ]

    # Load hyperopt trials object from each file
    loaded_trials = Trials()
    if previous_trials is not None:
        loaded_trials = merge_trials(loaded_trials, previous_trials.trials)

    for path in hyperopt_checkpoint_files:
        with open(path, "rb") as f:
            trial = pickle.load(f)
            loaded_trials = merge_trials(loaded_trials, trial.trials)

    return loaded_trials


def save_trials(
    dir_path: str, trials: Trials, hyperopt_seed: int, logger: logging.Logger = None
) -> None:
    """
    Saves hyperopt trial data as a `.pkl` file.

    :param dir_path: Path to the directory containing hyperopt checkpoint files.
    :param hyperopt_seed: The initial seed used for choosing parameters in hyperopt trials.
    :param trials: A trials object containing information on a completed hyperopt iteration.
    """
    if logger is None:
        info = print
    else:
        info = logger.info

    new_fname = f"{hyperopt_seed}.pkl"
    existing_files = os.listdir(dir_path)
    if new_fname in existing_files:
        info(
            f"When saving trial with unique seed {hyperopt_seed}, found that a trial with this seed already exists. "
            "This trial was not saved."
        )
    else:
        pickle.dump(trials, open(os.path.join(dir_path, new_fname), "wb"))


def get_hyperopt_seed(seed: int, dir_path: str) -> int:
    """
    Assigns a seed for hyperopt calculations. Each iteration will start with a different seed.

    :param seed: The initial attempted hyperopt seed.
    :param dir_path: Path to the directory containing hyperopt checkpoint files.
    :return: An integer for use as hyperopt random seed.
    """

    seed_path = os.path.join(dir_path, HYPEROPT_SEED_FILE_NAME)

    seeds = []
    if os.path.exists(seed_path):
        with open(seed_path, "r") as f:
            seed_line = next(f)
            seeds.extend(seed_line.split())
    else:
        makedirs(seed_path, isfile=True)

    seeds = [int(sd) for sd in seeds]

    while seed in seeds:
        seed += 1
    seeds.append(seed)

    write_line = " ".join(map(str, seeds)) + "\n"

    with open(seed_path, "w") as f:
        f.write(write_line)

    return seed


def load_manual_trials(
    manual_trials_dirs: List[str], param_keys: List[str], hyperopt_args: HyperoptArgs
) -> Trials:
    """
    Function for loading in manual training runs as trials for inclusion in hyperparameter search.
    Trials must be consistent with trials that would be generated in hyperparameter optimization.
    Parameters that are part of the search space do not have to match, but all others do.

    :param manual_trials_dirs: A list of paths to save directories for the manual trials, as would include test_scores.csv and args.json.
    :param param_keys: A list of the parameters included in the hyperparameter optimization.
    :param hyperopt_args: The arguments for the hyperparameter optimization job.
    :return: A hyperopt trials object including all the loaded manual trials.
    """
    # Non-extensive list of arguments that need to match between the manual trials and the search space.
    matching_args = [
        ("number_of_molecules", None),
        ("aggregation", "aggregation"),
        ("num_folds", None),
        ("ensemble_size", None),
        ("max_lr", "max_lr"),
        ("init_lr", "init_lr_ratio"),
        ("final_lr", "final_lr_ratio"),
        ("activation", "activation"),
        ("metric", None),
        ("bias", None),
        ("epochs", None),
        ("explicit_h", None),
        ("adding_h", None),
        ("reaction", None),
        ("split_type", None),
        ("warmup_epochs", "warmup_epochs"),
        ("aggregation_norm", "aggregation_norm"),
        ("batch_size", "batch_size"),
        ("depth", "depth"),
        ("dropout", "dropout"),
        ("ffn_num_layers", "ffn_num_layers"),
        ("dataset_type", None),
        ("multiclass_num_classes", None),
        ("features_generator", None),
        ("no_features_scaling", None),
        ("features_only", None),
        ("split_sizes", None),
    ]

    manual_trials_data = []
    for i, trial_dir in enumerate(manual_trials_dirs):

        # Extract trial data from test_scores.csv
        with open(os.path.join(trial_dir, "test_scores.csv")) as f:
            reader = csv.reader(f)
            next(reader)
            read_line = next(reader)
        mean_score = float(read_line[1])
        std_score = float(read_line[2])
        loss = (1 if hyperopt_args.minimize_score else -1) * mean_score

        # Extract argument data from args.json
        with open(os.path.join(trial_dir, "args.json")) as f:
            trial_args = json.load(f)

        # Check for differences in manual trials and hyperopt space
        if "linked_hidden_size" in param_keys:
            if trial_args["hidden_size"] != trial_args["ffn_hidden_size"]:
                raise ValueError(
                    f'The manual trial in {trial_dir} has a hidden_size {trial_args["hidden_size"]} '
                    f'that does not match its ffn_hidden_size {trial_args["ffn_hidden_size"]}, as it would in hyperparameter search.'
                )
        elif "hidden_size" not in param_keys or "ffn_hidden_size" not in param_keys:
            if "hidden_size" not in param_keys:
                if getattr(hyperopt_args, "hidden_size") != trial_args["hidden_size"]:
                    raise ValueError(
                        f"Manual trial {trial_dir} has different training argument hidden_size than the hyperparameter optimization search trials."
                    )
            if "ffn_hidden_size" not in param_keys:
                if (
                    getattr(hyperopt_args, "ffn_hidden_size")
                    != trial_args["ffn_hidden_size"]
                ):
                    raise ValueError(
                        f"Manual trial {trial_dir} has different training argument ffn_hidden_size than the hyperparameter optimization search trials."
                    )

        for arg, space_parameter in matching_args:
            if space_parameter not in param_keys:
                if getattr(hyperopt_args, arg) != trial_args[arg]:
                    raise ValueError(
                        f"Manual trial {trial_dir} has different training argument {arg} than the hyperparameter optimization search trials."
                    )

        # Construct data dict
        param_dict = {}
        vals_dict = {}
        for key in param_keys:
            if key == "init_lr_ratio":
                param_value = val_value = trial_args["init_lr"] / trial_args["max_lr"]
            elif key == "final_lr_ratio":
                param_value = val_value = trial_args["final_lr"] / trial_args["max_lr"]
            elif key == "linked_hidden_size":
                param_value = val_value = trial_args["hidden_size"]
            elif key == "aggregation":
                param_value = trial_args[key]
                val_value = ["mean", "sum", "norm"].index(param_value)
            elif key == "activation":
                param_value = trial_args[key]
                val_value = ["ReLU", "LeakyReLU", "PReLU", "tanh", "SELU", "ELU"].index(
                    param_value
                )
            else:
                param_value = val_value = trial_args[key]
            param_dict[key] = param_value
            vals_dict[key] = [val_value]
        idxs_dict = {key: [i] for key in param_keys}
        results_dict = {
            "loss": loss,
            "status": "ok",
            "mean_score": mean_score,
            "std_score": std_score,
            "hyperparams": param_dict,
            "num_params": 0,
            "seed": -(i + 1),
        }
        misc_dict = {
            "tid": i,
            "cmd": ("domain_attachment", "FMinIter_Domain"),
            "workdir": None,
            "idxs": idxs_dict,
            "vals": vals_dict,
        }
        trial_data = {
            "state": 2,
            "tid": i,
            "spec": None,
            "result": results_dict,
            "misc": misc_dict,
            "exp_key": None,
            "owner": None,
            "version": 0,
            "book_time": None,
            "refresh_time": None,
        }
        manual_trials_data.append(trial_data)

    trials = Trials()
    trials = merge_trials(trials=trials, new_trials_data=manual_trials_data)
    return trials


def save_config(config_path: str, hyperparams_dict: dict, max_lr: float) -> None:
    """
    Saves the hyperparameters for the best trial to a config json file.

    :param config_path: File path for the config json file.
    :param hyperparams_dict: A dictionary of hyperparameters found during the search.
    :param max_lr: The maximum learning rate value, to be used if not a search parameter.
    """
    makedirs(config_path, isfile=True)

    save_dict = {}

    for key in hyperparams_dict:
        if key == "linked_hidden_size":
            save_dict["hidden_size"] = hyperparams_dict["linked_hidden_size"]
            save_dict["ffn_hidden_size"] = hyperparams_dict["linked_hidden_size"]
        elif key == "init_lr_ratio":
            if "max_lr" not in hyperparams_dict:
                save_dict["init_lr"] = hyperparams_dict[key] * max_lr
            else:
                save_dict["init_lr"] = (
                    hyperparams_dict[key] * hyperparams_dict["max_lr"]
                )
        elif key == "final_lr_ratio":
            if "max_lr" not in hyperparams_dict:
                save_dict["final_lr"] = hyperparams_dict[key] * max_lr
            else:
                save_dict["final_lr"] = (
                    hyperparams_dict[key] * hyperparams_dict["max_lr"]
                )
        else:
            save_dict[key] = hyperparams_dict[key]

    with open(config_path, "w") as f:
        json.dump(save_dict, f, indent=4, sort_keys=True)



"""
Runs the web interface version of Chemprop.
Designed to be used for production only, along with Gunicorn.
"""
from chemprop.web.app import app, db
from chemprop.web.utils import clear_temp_folder, set_root_folder


def build_app(*args, **kwargs):
    # Set up root folder and subfolders
    set_root_folder(
        app=app,
        root_folder=kwargs.get('root_folder', None),
        create_folders=True
    )
    clear_temp_folder(app=app)

    db.init_app(app)
    if 'init_db' in kwargs:
        with app.app_context():
            db.init_db()
            print("-- INITIALIZED DATABASE --")

    app.config['DEMO'] = kwargs.get('demo', False)

    return app



import math
from typing import Callable, Dict, List, Set, Tuple

import numpy as np
from rdkit import Chem

from chemprop.args import InterpretArgs
from chemprop.data import get_data_from_smiles, get_header, get_smiles, MoleculeDataLoader, MoleculeDataset
from chemprop.train import predict
from chemprop.utils import load_args, load_checkpoint, load_scalers, timeit


MIN_ATOMS = 15
C_PUCT = 10


class ChempropModel:
    """A :class:`ChempropModel` is a wrapper around a :class:`~chemprop.models.model.MoleculeModel` for interpretation."""

    def __init__(self, args: InterpretArgs) -> None:
        """
        :param args: A :class:`~chemprop.args.InterpretArgs` object containing arguments for interpretation.
        """
        self.args = args
        self.train_args = load_args(args.checkpoint_paths[0])

        # If features were used during training, they must be used when predicting
        if ((self.train_args.features_path is not None or self.train_args.features_generator is not None)
                and args.features_generator is None):
            raise ValueError('Features were used during training so they must be specified again during prediction '
                             'using the same type of features as before (with --features_generator <generator> '
                             'and using --no_features_scaling if applicable).')

        if self.train_args.atom_descriptors_size > 0 or self.train_args.atom_features_size > 0 or self.train_args.bond_descriptors_size > 0 or self.train_args.bond_features_size > 0:
            raise NotImplementedError('The interpret function does not yet work with additional atom or bond features')

        self.scaler, self.features_scaler, self.atom_descriptor_scaler, self.bond_descriptor_scaler, self.atom_bond_scaler = load_scalers(args.checkpoint_paths[0])
        self.checkpoints = [load_checkpoint(checkpoint_path, device=args.device) for checkpoint_path in args.checkpoint_paths]

    def __call__(self, smiles: List[str], batch_size: int = 500) -> List[List[float]]:
        """
        Makes predictions on a list of SMILES.

        :param smiles: A list of SMILES to make predictions on.
        :param batch_size: The batch size.
        :return: A list of lists of floats containing the predicted values.
        """
        test_data = get_data_from_smiles(smiles=smiles, skip_invalid_smiles=False, features_generator=self.args.features_generator)
        valid_indices = [i for i in range(len(test_data)) if test_data[i].mol is not None]
        test_data = MoleculeDataset([test_data[i] for i in valid_indices])

        if self.train_args.features_scaling:
            test_data.normalize_features(self.features_scaler)
        if self.train_args.atom_descriptor_scaling and self.args.atom_descriptors is not None:
            test_data.normalize_features(self.atom_descriptor_scaler, scale_atom_descriptors=True)
        if self.train_args.bond_descriptor_scaling and self.args.bond_descriptors_size > 0:
            test_data.normalize_features(self.bond_descriptor_scaler, scale_bond_descriptors=True)

        test_data_loader = MoleculeDataLoader(dataset=test_data, batch_size=batch_size, num_workers=self.args.num_workers)

        sum_preds = []
        for model in self.checkpoints:
            model_preds = predict(
                model=model,
                data_loader=test_data_loader,
                scaler=self.scaler,
                disable_progress_bar=True
            )
            sum_preds.append(np.array(model_preds))

        # Ensemble predictions
        sum_preds = sum(sum_preds)
        avg_preds = sum_preds / len(self.checkpoints)

        return avg_preds


class MCTSNode:
    """A :class:`MCTSNode` represents a node in a Monte Carlo Tree Search."""

    def __init__(self, smiles: str, atoms: List[int], W: float = 0, N: int = 0, P: float = 0) -> None:
        """
        :param smiles: The SMILES for the substructure at this node.
        :param atoms: A list of atom indices represented by this node.
        :param W: The W value of this node.
        :param N: The N value of this node.
        :param P: The P value of this node.
        """
        self.smiles = smiles
        self.atoms = set(atoms)
        self.children = []
        self.W = W
        self.N = N
        self.P = P

    def Q(self) -> float:
        return self.W / self.N if self.N > 0 else 0

    def U(self, n: int) -> float:
        return C_PUCT * self.P * math.sqrt(n) / (1 + self.N)


def find_clusters(mol: Chem.Mol) -> Tuple[List[Tuple[int, ...]], List[List[int]]]:
    """
    Finds clusters within the molecule.

    :param mol: An RDKit molecule.
    :return: A tuple containing a list of atom tuples representing the clusters
             and a list of lists of atoms in each cluster.
    """
    n_atoms = mol.GetNumAtoms()
    if n_atoms == 1:  # special case
        return [(0,)], [[0]]

    clusters = []
    for bond in mol.GetBonds():
        a1 = bond.GetBeginAtom().GetIdx()
        a2 = bond.GetEndAtom().GetIdx()
        if not bond.IsInRing():
            clusters.append((a1, a2))

    ssr = [tuple(x) for x in Chem.GetSymmSSSR(mol)]
    clusters.extend(ssr)

    atom_cls = [[] for _ in range(n_atoms)]
    for i in range(len(clusters)):
        for atom in clusters[i]:
            atom_cls[atom].append(i)

    return clusters, atom_cls


def __extract_subgraph(mol: Chem.Mol, selected_atoms: Set[int]) -> Tuple[Chem.Mol, List[int]]:
    """
    Extracts a subgraph from an RDKit molecule given a set of atom indices.

    :param mol: An RDKit molecule from which to extract a subgraph.
    :param selected_atoms: The atoms which form the subgraph to be extracted.
    :return: A tuple containing an RDKit molecule representing the subgraph
             and a list of root atom indices from the selected indices.
    """
    selected_atoms = set(selected_atoms)
    roots = []
    for idx in selected_atoms:
        atom = mol.GetAtomWithIdx(idx)
        bad_neis = [y for y in atom.GetNeighbors() if y.GetIdx() not in selected_atoms]
        if len(bad_neis) > 0:
            roots.append(idx)

    new_mol = Chem.RWMol(mol)

    for atom_idx in roots:
        atom = new_mol.GetAtomWithIdx(atom_idx)
        atom.SetAtomMapNum(1)
        aroma_bonds = [bond for bond in atom.GetBonds() if bond.GetBondType() == Chem.rdchem.BondType.AROMATIC]
        aroma_bonds = [bond for bond in aroma_bonds if
                       bond.GetBeginAtom().GetIdx() in selected_atoms and bond.GetEndAtom().GetIdx() in selected_atoms]
        if len(aroma_bonds) == 0:
            atom.SetIsAromatic(False)

    remove_atoms = [atom.GetIdx() for atom in new_mol.GetAtoms() if atom.GetIdx() not in selected_atoms]
    remove_atoms = sorted(remove_atoms, reverse=True)
    for atom in remove_atoms:
        new_mol.RemoveAtom(atom)

    return new_mol.GetMol(), roots


def extract_subgraph(smiles: str, selected_atoms: Set[int]) -> Tuple[str, List[int]]:
    """
    Extracts a subgraph from a SMILES given a set of atom indices.

    :param smiles: A SMILES from which to extract a subgraph.
    :param selected_atoms: The atoms which form the subgraph to be extracted.
    :return: A tuple containing a SMILES representing the subgraph
             and a list of root atom indices from the selected indices.
    """
    # try with kekulization
    mol = Chem.MolFromSmiles(smiles)
    Chem.Kekulize(mol)
    subgraph, roots = __extract_subgraph(mol, selected_atoms)
    try:
        subgraph = Chem.MolToSmiles(subgraph, kekuleSmiles=True)
        subgraph = Chem.MolFromSmiles(subgraph)
    except Exception:
        subgraph = None

    mol = Chem.MolFromSmiles(smiles)  # de-kekulize
    if subgraph is not None and mol.HasSubstructMatch(subgraph):
        return Chem.MolToSmiles(subgraph), roots

    # If fails, try without kekulization
    subgraph, roots = __extract_subgraph(mol, selected_atoms)
    subgraph = Chem.MolToSmiles(subgraph)
    subgraph = Chem.MolFromSmiles(subgraph)

    if subgraph is not None:
        return Chem.MolToSmiles(subgraph), roots
    else:
        return None, None


def mcts_rollout(node: MCTSNode,
                 state_map: Dict[str, MCTSNode],
                 orig_smiles: str,
                 clusters: List[Set[int]],
                 atom_cls: List[Set[int]],
                 nei_cls: List[Set[int]],
                 scoring_function: Callable[[List[str]], List[float]]) -> float:
    """
    A Monte Carlo Tree Search rollout from a given :class:`MCTSNode`.

    :param node: The :class:`MCTSNode` from which to begin the rollout.
    :param state_map: A mapping from SMILES to :class:`MCTSNode`.
    :param orig_smiles: The original SMILES of the molecule.
    :param clusters: Clusters of atoms.
    :param atom_cls: Atom indices in the clusters.
    :param nei_cls: Neighboring clusters.
    :param scoring_function: A function for scoring subgraph SMILES using a Chemprop model.
    :return: The score of this MCTS rollout.
    """
    cur_atoms = node.atoms
    if len(cur_atoms) <= MIN_ATOMS:
        return node.P

    # Expand if this node has never been visited
    if len(node.children) == 0:
        cur_cls = set([i for i, x in enumerate(clusters) if x <= cur_atoms])
        for i in cur_cls:
            leaf_atoms = [a for a in clusters[i] if len(atom_cls[a] & cur_cls) == 1]
            if len(nei_cls[i] & cur_cls) == 1 or len(clusters[i]) == 2 and len(leaf_atoms) == 1:
                new_atoms = cur_atoms - set(leaf_atoms)
                new_smiles, _ = extract_subgraph(orig_smiles, new_atoms)
                if new_smiles in state_map:
                    new_node = state_map[new_smiles]  # merge identical states
                else:
                    new_node = MCTSNode(new_smiles, new_atoms)
                if new_smiles:
                    node.children.append(new_node)

        state_map[node.smiles] = node
        if len(node.children) == 0:
            return node.P  # cannot find leaves

        scores = scoring_function([[x.smiles] for x in node.children])
        for child, score in zip(node.children, scores):
            child.P = score

    sum_count = sum(c.N for c in node.children)
    selected_node = max(node.children, key=lambda x: x.Q() + x.U(sum_count))
    v = mcts_rollout(selected_node, state_map, orig_smiles, clusters, atom_cls, nei_cls, scoring_function)
    selected_node.W += v
    selected_node.N += 1

    return v


def mcts(smiles: str,
         scoring_function: Callable[[List[str]], List[float]],
         n_rollout: int,
         max_atoms: int,
         prop_delta: float) -> List[MCTSNode]:
    """
    Runs the Monte Carlo Tree Search algorithm.

    :param smiles: The SMILES of the molecule to perform the search on.
    :param scoring_function: A function for scoring subgraph SMILES using a Chemprop model.
    :param n_rollout: THe number of MCTS rollouts to perform.
    :param max_atoms: The maximum number of atoms allowed in an extracted rationale.
    :param prop_delta: The minimum required property value for a satisfactory rationale.
    :return: A list of rationales each represented by a :class:`MCTSNode`.
    """
            
    mol = Chem.MolFromSmiles(smiles)
    if mol.GetNumAtoms() > 50:
        n_rollout = 1

    clusters, atom_cls = find_clusters(mol)
    nei_cls = [0] * len(clusters)
    for i, cls in enumerate(clusters):
        nei_cls[i] = [nei for atom in cls for nei in atom_cls[atom]]
        nei_cls[i] = set(nei_cls[i]) - {i}
        clusters[i] = set(list(cls))
    for a in range(len(atom_cls)):
        atom_cls[a] = set(atom_cls[a])

    root = MCTSNode(smiles, set(range(mol.GetNumAtoms())))
    state_map = {smiles: root}
    for _ in range(n_rollout):
        mcts_rollout(root, state_map, smiles, clusters, atom_cls, nei_cls, scoring_function)

    rationales = [node for _, node in state_map.items() if len(node.atoms) <= max_atoms and node.P >= prop_delta]

    return rationales


@timeit()
def interpret(args: InterpretArgs) -> None:
    """
    Runs interpretation of a Chemprop model using the Monte Carlo Tree Search algorithm.

    :param args: A :class:`~chemprop.args.InterpretArgs` object containing arguments for interpretation.
    """

    if args.number_of_molecules != 1:
        raise ValueError(
            "Interpreting is currently only available for single-molecule models."
        )

    global C_PUCT, MIN_ATOMS

    chemprop_model = ChempropModel(args)

    def scoring_function(smiles: List[str]) -> List[float]:
        return chemprop_model(smiles)[:, args.property_id - 1]

    C_PUCT = args.c_puct
    MIN_ATOMS = args.min_atoms

    all_smiles = get_smiles(path=args.data_path, smiles_columns=args.smiles_columns)
    header = get_header(path=args.data_path)

    property_name = header[args.property_id] if len(header) > args.property_id else 'score'
    print(f'smiles,{property_name},rationale,rationale_score')

    for smiles in all_smiles:
        score = scoring_function([smiles])[0]
        if score > args.prop_delta:
            rationales = mcts(
                smiles=smiles[0],
                scoring_function=scoring_function,
                n_rollout=args.rollout,
                max_atoms=args.max_atoms,
                prop_delta=args.prop_delta
            )
        else:
            rationales = []

        if len(rationales) == 0:
            print(f'{smiles},{score:.3f},,')
        else:
            min_size = min(len(x.atoms) for x in rationales)
            min_rationales = [x for x in rationales if len(x.atoms) == min_size]
            rats = sorted(min_rationales, key=lambda x: x.P, reverse=True)
            print(f'{smiles},{score:.3f},{rats[0].smiles},{rats[0].P:.3f}')


def chemprop_interpret() -> None:
    """Runs interpretation of a Chemprop model.

    This is the entry point for the command line command :code:`chemprop_interpret`.
    """
    interpret(args=InterpretArgs().parse_args())



from abc import ABC, abstractmethod
from typing import List

import numpy as np
from scipy.stats import t, spearmanr
from scipy.special import erfinv

from chemprop.uncertainty.uncertainty_calibrator import UncertaintyCalibrator
from chemprop.train import evaluate_predictions


class UncertaintyEvaluator(ABC):
    """
    A class for evaluating the effectiveness of uncertainty estimates with metrics.
    """

    def __init__(
        self,
        evaluation_method: str,
        calibration_method: str,
        uncertainty_method: str,
        dataset_type: str,
        loss_function: str,
        calibrator: UncertaintyCalibrator,
        is_atom_bond_targets: bool,
    ):
        self.evaluation_method = evaluation_method
        self.calibration_method = calibration_method
        self.uncertainty_method = uncertainty_method
        self.dataset_type = dataset_type
        self.loss_function = loss_function
        self.calibrator = calibrator
        self.is_atom_bond_targets = is_atom_bond_targets

        self.raise_argument_errors()

    def raise_argument_errors(self):
        """
        Raise errors for incompatibilities between dataset type and uncertainty method, or similar.
        """
        if self.dataset_type == "spectra":
            raise ValueError(
                "No uncertainty evaluators implemented for spectra dataset type."
            )
        if self.uncertainty_method in ["ensemble", "dropout"] and self.dataset_type in [
            "classification",
            "multiclass",
        ]:
            raise ValueError(
                "Though ensemble and dropout uncertainty methods are available for classification \
                    multiclass dataset types, their outputs are not confidences and are not \
                    compatible with any implemented evaluation methods for classification."
            )
        if self.uncertainty_method == "dirichlet":
            raise ValueError(
                "The Dirichlet uncertainty method returns an evidential uncertainty value rather than a \
                    class confidence. It is not compatible with any implemented evaluation methods. \
                    To evaluate the performance of a model trained using the Dirichlet loss function, \
                    use the classification uncertainty method in a separate job."
            )

    @abstractmethod
    def evaluate(
        self,
        targets: List[List[float]],
        preds: List[List[float]],
        uncertainties: List[List[float]],
        mask: List[List[bool]],
    ) -> List[float]:
        """
        Evaluate the performance of uncertainty predictions against the model target values.

        :param targets:  The target values for prediction.
        :param preds: The prediction values of a model on the test set.
        :param uncertainties: The estimated uncertainty values, either calibrated or uncalibrated, of a model on the test set.
        :param mask: Whether the values in targets were provided.

        :return: A list of metric values for each model task.
        """


class MetricEvaluator(UncertaintyEvaluator):
    """
    A class for evaluating confidence estimates of classification and multiclass datasets using builtin evaluation metrics.
    """

    def evaluate(
        self,
        targets: List[List[float]],
        preds: List[List[float]],
        uncertainties: List[List[float]],
        mask: List[List[bool]],
    ):
        return evaluate_predictions(
            preds=uncertainties,
            targets=targets,
            num_tasks=np.array(targets).shape[1],
            metrics=[self.evaluation_method],
            dataset_type=self.dataset_type,
        )[self.evaluation_method]


class NLLRegressionEvaluator(UncertaintyEvaluator):
    """
    A class for evaluating regression uncertainty values using the mean negative-log-likelihood
    of the actual targets given the probability distributions estimated by the model.
    """

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "regression":
            raise ValueError(
                "NLL Regression Evaluator is only for regression dataset types."
            )

    def evaluate(
        self,
        targets: List[List[float]],
        preds: List[List[float]],
        uncertainties: List[List[float]],
        mask: List[List[bool]],
    ):
        if self.calibrator is None:  # uncalibrated regression uncertainties are variances
            uncertainties = np.array(uncertainties)
            preds = np.array(preds)
            targets = np.array(targets)
            mask = np.array(mask)
            num_tasks = len(mask)
            if self.is_atom_bond_targets:
                uncertainties = [np.concatenate(x) for x in zip(*uncertainties)]
                preds = [np.concatenate(x) for x in zip(*preds)]
                targets = [np.concatenate(x) for x in zip(*targets)]
            else:
                uncertainties = np.array(list(zip(*uncertainties)))
                preds = np.array(list(zip(*preds)))
                targets = targets.astype(float)
                targets = np.array(list(zip(*targets)))
            nll = []
            for i in range(num_tasks):
                task_mask = mask[i]
                task_unc = uncertainties[i][task_mask]
                task_preds = preds[i][task_mask]
                task_targets = targets[i][task_mask]
                task_nll = np.log(2 * np.pi * task_unc) / 2 \
                    + (task_preds - task_targets) ** 2 / (2 * task_unc)
                nll.append(task_nll.mean())
            return nll
        else:
            nll = self.calibrator.nll(
                preds=preds, unc=uncertainties, targets=targets, mask=mask
            )  # shape(data, task)
            return nll


class NLLClassEvaluator(UncertaintyEvaluator):
    """
    A class for evaluating classification uncertainty values using the mean negative-log-likelihood
    of the actual targets given the probabilities assigned to them by the model.
    """

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "classification":
            raise ValueError(
                "NLL Classification Evaluator is only for classification dataset types."
            )

    def evaluate(
        self,
        targets: List[List[float]],
        preds: List[List[float]],
        uncertainties: List[List[float]],
        mask: List[List[bool]],
    ):
        targets = np.array(targets)
        mask = np.array(mask)
        num_tasks = len(mask)
        uncertainties = np.array(uncertainties)
        if self.is_atom_bond_targets:
            uncertainties = [np.concatenate(x) for x in zip(*uncertainties)]
            targets = [np.concatenate(x) for x in zip(*targets)]
        else:
            uncertainties = np.array(list(zip(*uncertainties)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
        nll = []
        for i in range(num_tasks):
            task_mask = mask[i]
            task_unc = uncertainties[i][task_mask]
            task_targets = targets[i][task_mask]
            task_likelihood = task_unc * task_targets + (1 - task_unc) * (1 - task_targets)
            task_nll = -1 * np.log(task_likelihood)
            nll.append(task_nll.mean())
        return nll


class NLLMultiEvaluator(UncertaintyEvaluator):
    """
    A class for evaluating multiclass uncertainty values using the mean negative-log-likelihood
    of the actual targets given the probabilities assigned to them by the model.
    """

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "multiclass":
            raise ValueError(
                "NLL Multiclass Evaluator is only for multiclass dataset types."
            )

    def evaluate(
        self,
        targets: List[List[float]],
        preds: List[List[float]],
        uncertainties: List[List[float]],
        mask: List[List[bool]],
    ):
        targets = np.array(targets, dtype=int)  # shape(data, tasks)
        mask = np.array(mask)
        num_tasks = len(mask)
        uncertainties = np.array(uncertainties)
        nll = []
        for i in range(num_tasks):
            task_mask = mask[:, i]
            task_preds = uncertainties[task_mask, i]
            task_targets = targets[task_mask, i]  # shape(data)
            bin_targets = np.zeros_like(task_preds)  # shape(data, classes)
            bin_targets[np.arange(task_targets.shape[0]), task_targets] = 1
            task_likelihood = np.sum(bin_targets * task_preds, axis=1)
            task_nll = -1 * np.log(task_likelihood)
            nll.append(task_nll.mean())
        return nll


class CalibrationAreaEvaluator(UncertaintyEvaluator):
    """
    A class for evaluating regression uncertainty values based on how they deviate from perfect
    calibration on an observed-probability versus expected-probability plot.
    """

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "regression":
            raise ValueError(
                "Miscalibration area is only implemented for regression dataset types."
            )

    def evaluate(
        self,
        targets: List[List[float]],
        preds: List[List[float]],
        uncertainties: List[List[float]],
        mask: List[List[bool]],
    ):
        targets = np.array(targets)  # shape(data, tasks)
        mask = np.array(mask)
        num_tasks = len(mask)
        uncertainties = np.array(uncertainties)
        preds = np.array(preds)

        if self.is_atom_bond_targets:
            uncertainties = [np.concatenate(x) for x in zip(*uncertainties)]
            targets = [np.concatenate(x) for x in zip(*targets)]
            preds = [np.concatenate(x) for x in zip(*preds)]
        else:
            uncertainties = np.array(list(zip(*uncertainties)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
            preds = np.array(list(zip(*preds)))
        # using 101 bin edges, hardcoded
        fractions = np.zeros([num_tasks, 101])  # shape(tasks, 101)
        fractions[:, 100] = 1

        if self.calibrator is not None:
            original_metric = self.calibrator.regression_calibrator_metric
            original_scaling = self.calibrator.scaling
            original_interval = self.calibrator.interval_percentile

            bin_scaling = [0]

            for i in range(1, 100):
                self.calibrator.regression_calibrator_metric = "interval"
                self.calibrator.interval_percentile = i
                self.calibrator.calibrate()
                bin_scaling.append(self.calibrator.scaling)

            for j in range(num_tasks):
                task_mask = mask[j]
                task_targets = targets[j][task_mask]
                task_preds = preds[j][task_mask]
                task_error = np.abs(task_preds - task_targets)
                task_unc = uncertainties[j][task_mask]

                for i in range(1, 100):
                    bin_unc = task_unc / original_scaling[j] * bin_scaling[i][j]
                    bin_fraction = np.mean(bin_unc >= task_error)
                    fractions[j, i] = bin_fraction

            # return calibration settings to original state
            self.calibrator.regression_calibrator_metric = original_metric
            self.calibrator.scaling = original_scaling
            self.calibrator.interval_percentile = original_interval

        else:  # uncertainties are uncalibrated variances
            bin_scaling = [0]
            for i in range(1, 100):
                bin_scaling.append(erfinv(i / 100) * np.sqrt(2))
            for j in range(num_tasks):
                task_mask = mask[j]
                task_targets = targets[j][task_mask]
                task_preds = preds[j][task_mask]
                task_error = np.abs(task_preds - task_targets)
                task_unc = uncertainties[j][task_mask]
                for i in range(1, 100):
                    bin_unc = np.sqrt(task_unc) * bin_scaling[i]
                    bin_fraction = np.mean(bin_unc >= task_error)
                    fractions[j, i] = bin_fraction

        # trapezoid rule
        auce = np.sum(
            0.01 * np.abs(fractions - np.expand_dims(np.arange(101) / 100, axis=0)),
            axis=1,
        )
        return auce.tolist()


class ExpectedNormalizedErrorEvaluator(UncertaintyEvaluator):
    """
    A class that evaluates uncertainty performance by binning together clusters of predictions
    and comparing the average predicted variance of the clusters against the RMSE of the cluster.
    Method discussed in https://doi.org/10.1021/acs.jcim.9b00975.
    """

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "regression":
            raise ValueError(
                "Expected normalized error is only appropriate for regression dataset types."
            )

    def evaluate(
        self,
        targets: List[List[float]],
        preds: List[List[float]],
        uncertainties: List[List[float]],
        mask: List[List[bool]],
    ):
        targets = np.array(targets)  # shape(data, tasks)
        mask = np.array(mask)
        num_tasks = len(mask)
        uncertainties = np.array(uncertainties)
        preds = np.array(preds)
        if self.is_atom_bond_targets:
            uncertainties = [np.concatenate(x) for x in zip(*uncertainties)]
            targets = [np.concatenate(x) for x in zip(*targets)]
            preds = [np.concatenate(x) for x in zip(*preds)]
        else:
            uncertainties = np.array(list(zip(*uncertainties)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
            preds = np.array(list(zip(*preds)))
        # get stdev scaling then revert if interval
        if self.calibrator is not None:
            original_metric = self.calibrator.regression_calibrator_metric
            original_scaling = self.calibrator.scaling
            if (
                self.calibration_method != "tscaling"
                and self.calibrator.regression_calibrator_metric == "interval"
            ):
                self.calibrator.regression_calibrator_metric = "stdev"
                self.calibrator.calibrate()
                stdev_scaling = self.calibrator.scaling
                self.calibrator.regression_calibrator_metric = original_metric
                self.calibrator.scaling = original_scaling

        root_mean_vars = np.zeros([num_tasks, 100])  # shape(tasks, 100)
        rmses = np.zeros_like(root_mean_vars)

        for i in range(num_tasks):
            task_mask = mask[i]  # shape(data)
            task_targets = targets[i][task_mask]
            task_preds = preds[i][task_mask]
            task_error = np.abs(task_preds - task_targets)
            task_unc = uncertainties[i][task_mask]

            sort_idx = np.argsort(task_unc)
            task_unc = task_unc[sort_idx]
            task_error = task_error[sort_idx]

            # 100 bins
            split_unc = np.array_split(task_unc, 100)  # shape(list100, data/100)
            split_error = np.array_split(task_error, 100)

            for j in range(100):
                if self.calibrator is None:  # starts as a variance
                    root_mean_vars[i, j] = np.sqrt(np.mean(split_unc[j]))
                    rmses[i, j] = np.sqrt(np.mean(np.square(split_error[j])))
                elif self.calibration_method == "tscaling":  # convert back to sample stdev
                    bin_unc = split_unc[j] / original_scaling[i]
                    bin_var = t.var(df=self.calibrator.num_models - 1, scale=bin_unc)
                    root_mean_vars[i, j] = np.sqrt(np.mean(bin_var))
                    rmses[i, j] = np.sqrt(np.mean(np.square(split_error[j])))
                else:
                    bin_unc = split_unc[j]
                    if self.calibrator.regression_calibrator_metric == "interval":
                        bin_unc = bin_unc / original_scaling[i] * stdev_scaling[i]  # convert from interval to stdev as needed
                    root_mean_vars[i, j] = np.sqrt(np.mean(np.square(bin_unc)))
                    rmses[i, j] = np.sqrt(np.mean(np.square(split_error[j])))

        ence = np.mean(np.abs(root_mean_vars - rmses) / root_mean_vars, axis=1)
        return ence.tolist()


class SpearmanEvaluator(UncertaintyEvaluator):
    """
    Class evaluating uncertainty performance using the spearman rank correlation. Method produces
    better scores (closer to 1 in the [-1, 1] range) when the uncertainty values are predictive
    of the ranking of prediciton errors.
    """

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "regression":
            raise ValueError(
                "Spearman rank correlation is only appropriate for regression dataset types."
            )

    def evaluate(
        self,
        targets: List[List[float]],
        preds: List[List[float]],
        uncertainties: List[List[float]],
        mask: List[List[bool]],
    ):
        targets = np.array(targets)  # shape(data, tasks)
        uncertainties = np.array(uncertainties)
        mask = np.array(mask)
        num_tasks = len(mask)
        preds = np.array(preds)
        spearman_coeffs = []
        if self.is_atom_bond_targets:
            uncertainties = [np.concatenate(x) for x in zip(*uncertainties)]
            targets = [np.concatenate(x) for x in zip(*targets)]
            preds = [np.concatenate(x) for x in zip(*preds)]
        else:
            uncertainties = np.array(list(zip(*uncertainties)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
            preds = np.array(list(zip(*preds)))
        for i in range(num_tasks):
            task_mask = mask[i]
            task_unc = uncertainties[i][task_mask]
            task_targets = targets[i][task_mask]
            task_preds = preds[i][task_mask]
            task_error = np.abs(task_preds - task_targets)
            spmn = spearmanr(task_unc, task_error).correlation
            spearman_coeffs.append(spmn)
        return spearman_coeffs


class ConformalRegressionEvaluator(UncertaintyEvaluator):
    """
    A class for evaluating the coverage of conformal regression intervals.
    """

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "regression":
            raise ValueError(
                "Conformal Regression Evaluator is only for regression dataset types."
            )

    def evaluate(
        self,
        targets: List[List[float]],
        preds: List[List[float]],
        uncertainties: List[List[float]],
        mask: List[List[bool]],
    ):
        """
        Args:
            targets: shape(data, tasks)
            preds: shape(data, tasks)
            uncertainties: shape(data, tasks)
            mask: shape(data, tasks)

        Returns:
            Conformal coverage for each task
        """
        uncertainties = np.array(uncertainties)
        targets = np.array(targets)
        preds = np.array(preds)
        mask = np.array(mask)
        num_tasks = uncertainties.shape[1]
        if self.is_atom_bond_targets:
            uncertainties = [np.concatenate(x) for x in zip(*uncertainties)]
            targets = [np.concatenate(x) for x in zip(*targets)]
            preds = [np.concatenate(x) for x in zip(*preds)]
        else:
            uncertainties = np.array(list(zip(*uncertainties)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
            preds = np.array(list(zip(*preds)))

        results = []
        for i in range(num_tasks):
            task_mask = mask[i]
            task_unc = uncertainties[i][task_mask]
            task_targets = targets[i][task_mask]
            task_preds = preds[i][task_mask]
            unc_task_lower = task_preds - task_unc
            unc_task_upper = task_preds + task_unc
            task_results = np.logical_and(unc_task_lower <= task_targets, task_targets <= unc_task_upper)
            results.append(task_results.sum() / task_results.shape[0])

        return results


class ConformalMulticlassEvaluator(UncertaintyEvaluator):
    """
    A class for evaluating the coverage of conformal prediction on multiclass datasets.
    """

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "multiclass":
            raise ValueError(
                "Conformal Multiclass Evaluator is only for multiclass dataset types."
            )

    def evaluate(
        self,
        targets: List[List[float]],
        preds: List[List[float]],
        uncertainties: List[List[float]],
        mask: List[List[bool]],
    ):
        """
        Args:
            targets: shape(data, tasks)
            preds: shape(data, tasks, num_classes)
            uncertainties: shape(data, tasks, num_classes)
            mask: shape(data, tasks)

        Returns:
            Conformal coverage for each task
        """
        targets = np.array(targets, dtype=float)
        mask = np.array(mask, dtype=bool)
        uncertainties = np.array(uncertainties)
        num_tasks = targets.shape[1]
        results = []

        for i in range(num_tasks):
            task_mask = mask[i]
            task_results = np.take_along_axis(
                uncertainties[task_mask, i], targets[task_mask, i].reshape(-1, 1).astype(int), axis=1
            ).squeeze(1)
            results.append(task_results.sum() / len(task_results))

        return results


class ConformalMultilabelEvaluator(UncertaintyEvaluator):
    """
    A class for evaluating the coverage of conformal prediction on multilabel datasets.
    """

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "classification":
            raise ValueError(
                "Conformal Multilabel Evaluator is only for classification dataset types."
            )

    def evaluate(
        self,
        targets: List[List[float]],
        preds: List[List[float]],
        uncertainties: List[List[float]],
        mask: List[List[bool]],
    ):
        """
        Args:
            targets: shape(data, tasks)
            preds: shape(data, tasks)
            uncertainties: shape(data, tasks)
            mask: shape(data, tasks)

        Returns:
            Conformal coverage for each task
        """
        targets = np.array(targets, dtype=float)
        uncertainties = np.array(uncertainties)
        mask = np.array(mask)
        num_tasks = len(mask)
        if self.is_atom_bond_targets:
            uncertainties = [np.concatenate(x) for x in zip(*uncertainties)]
            targets = [np.concatenate(x) for x in zip(*targets)]
        else:
            uncertainties = np.array(list(zip(*uncertainties)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
        results = []
        for i in range(num_tasks):
            task_mask = mask[i]
            task_targets = targets[i][task_mask]
            task_unc_in = uncertainties[i][task_mask]
            task_unc_out = uncertainties[i + num_tasks][task_mask]
            task_results = np.logical_and(task_unc_in <= task_targets, task_targets <= task_unc_out)
            results.append(task_results.sum() / task_results.shape[0])

        return results


def build_uncertainty_evaluator(
    evaluation_method: str,
    calibration_method: str,
    uncertainty_method: str,
    dataset_type: str,
    loss_function: str,
    calibrator: UncertaintyCalibrator,
    is_atom_bond_targets: bool,
) -> UncertaintyEvaluator:
    """
    Function that chooses and returns the appropriate :class: `UncertaintyEvaluator` subclass
    for the provided arguments.
    """
    supported_evaluators = {
        "nll": {
            "regression": NLLRegressionEvaluator,
            "classification": NLLClassEvaluator,
            "multiclass": NLLMultiEvaluator,
            "spectra": None,
        }[dataset_type],
        "miscalibration_area": CalibrationAreaEvaluator,
        "ence": ExpectedNormalizedErrorEvaluator,
        "spearman": SpearmanEvaluator,
        "conformal_coverage": {
            "regression": ConformalRegressionEvaluator,
            "multiclass": ConformalMulticlassEvaluator,
            "classification": ConformalMultilabelEvaluator,
        }[dataset_type],
    }

    classification_metrics = [
        "auc",
        "prc-auc",
        "accuracy",
        "binary_cross_entropy",
        "f1",
        "mcc",
    ]
    multiclass_metrics = ["cross_entropy", "accuracy", "f1", "mcc"]
    if dataset_type == "classification" and evaluation_method in classification_metrics:
        evaluator_class = MetricEvaluator
    elif dataset_type == "multiclass" and evaluation_method in multiclass_metrics:
        evaluator_class = MetricEvaluator
    else:
        evaluator_class = supported_evaluators.get(evaluation_method, None)

    if evaluator_class is None:
        raise NotImplementedError(
            f"Evaluator type {evaluation_method} is not supported. Available options are all calibration/multiclass metrics and {list(supported_evaluators.keys())}"
        )
    else:
        evaluator = evaluator_class(
            evaluation_method=evaluation_method,
            calibration_method=calibration_method,
            uncertainty_method=uncertainty_method,
            dataset_type=dataset_type,
            loss_function=loss_function,
            calibrator=calibrator,
            is_atom_bond_targets=is_atom_bond_targets,
        )
        return evaluator



import logging
from typing import Callable

import numpy as np
from tensorboardX import SummaryWriter
import torch
import torch.nn as nn
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
from tqdm import tqdm

from chemprop.args import TrainArgs
from chemprop.data import MoleculeDataLoader, MoleculeDataset, AtomBondScaler
from chemprop.models import MoleculeModel
from chemprop.nn_utils import compute_gnorm, compute_pnorm, NoamLR


def train(
    model: MoleculeModel,
    data_loader: MoleculeDataLoader,
    loss_func: Callable,
    optimizer: Optimizer,
    scheduler: _LRScheduler,
    args: TrainArgs,
    n_iter: int = 0,
    atom_bond_scaler: AtomBondScaler = None,
    logger: logging.Logger = None,
    writer: SummaryWriter = None,
) -> int:
    """
    Trains a model for an epoch.

    :param model: A :class:`~chemprop.models.model.MoleculeModel`.
    :param data_loader: A :class:`~chemprop.data.data.MoleculeDataLoader`.
    :param loss_func: Loss function.
    :param optimizer: An optimizer.
    :param scheduler: A learning rate scheduler.
    :param args: A :class:`~chemprop.args.TrainArgs` object containing arguments for training the model.
    :param n_iter: The number of iterations (training examples) trained on so far.
    :param atom_bond_scaler: A :class:`~chemprop.data.scaler.AtomBondScaler` fitted on the atomic/bond targets.
    :param logger: A logger for recording output.
    :param writer: A tensorboardX SummaryWriter.
    :return: The total number of iterations (training examples) trained on so far.
    """
    debug = logger.debug if logger is not None else print

    model.train()
    if model.is_atom_bond_targets:
        loss_sum, iter_count = [0]*(len(args.atom_targets) + len(args.bond_targets)), 0
    else:
        loss_sum = iter_count = 0

    for batch in tqdm(data_loader, total=len(data_loader), leave=False):
        # Prepare batch
        batch: MoleculeDataset
        mol_batch, features_batch, target_batch, mask_batch, atom_descriptors_batch, atom_features_batch, bond_descriptors_batch, bond_features_batch, constraints_batch, data_weights_batch = \
            batch.batch_graph(), batch.features(), batch.targets(), batch.mask(), batch.atom_descriptors(), \
            batch.atom_features(), batch.bond_descriptors(), batch.bond_features(), batch.constraints(), batch.data_weights()

        if model.is_atom_bond_targets:
            targets = []
            for dt in zip(*target_batch):
                dt = np.concatenate(dt)
                targets.append(torch.tensor([0 if x is None else x for x in dt], dtype=torch.float))
            masks = [torch.tensor(mask, dtype=torch.bool) for mask in mask_batch]
            if args.target_weights is not None:
                target_weights = [torch.ones(1, 1) * i for i in args.target_weights]  # shape(tasks, 1)
            else:
                target_weights = [torch.ones(1, 1) for i in targets]
            data_weights = batch.atom_bond_data_weights()
            data_weights = [torch.tensor(x).unsqueeze(1) for x in data_weights]

            natoms, nbonds = batch.number_of_atoms, batch.number_of_bonds
            natoms, nbonds = np.array(natoms).flatten(), np.array(nbonds).flatten()
            constraints_batch = np.transpose(constraints_batch).tolist()
            ind = 0
            for i in range(len(args.atom_targets)):
                if not args.atom_constraints[i]:
                    constraints_batch[ind] = None
                else:
                    mean, std = atom_bond_scaler.means[ind][0], atom_bond_scaler.stds[ind][0]
                    for j, natom in enumerate(natoms):
                        constraints_batch[ind][j] = (constraints_batch[ind][j] - natom * mean) / std
                    constraints_batch[ind] = torch.tensor(constraints_batch[ind]).to(args.device)
                ind += 1
            for i in range(len(args.bond_targets)):
                if not args.bond_constraints[i]:
                    constraints_batch[ind] = None
                else:
                    mean, std = atom_bond_scaler.means[ind][0], atom_bond_scaler.stds[ind][0]
                    for j, nbond in enumerate(nbonds):
                        constraints_batch[ind][j] = (constraints_batch[ind][j] - nbond * mean) / std
                    constraints_batch[ind] = torch.tensor(constraints_batch[ind]).to(args.device)
                ind += 1
            bond_types_batch = []
            for i in range(len(args.atom_targets)):
                bond_types_batch.append(None)
            for i in range(len(args.bond_targets)):
                if args.adding_bond_types and atom_bond_scaler is not None:
                    mean, std = atom_bond_scaler.means[i+len(args.atom_targets)][0], atom_bond_scaler.stds[i+len(args.atom_targets)][0]
                    bond_types = [(b.GetBondTypeAsDouble() - mean) / std for d in batch for b in d.mol[0].GetBonds()]
                    bond_types = torch.FloatTensor(bond_types).to(args.device)
                    bond_types_batch.append(bond_types)
                else:
                    bond_types_batch.append(None)
        else:
            mask_batch = np.transpose(mask_batch).tolist()
            masks = torch.tensor(mask_batch, dtype=torch.bool)  # shape(batch, tasks)
            targets = torch.tensor([[0 if x is None else x for x in tb] for tb in target_batch])  # shape(batch, tasks)

            if args.target_weights is not None:
                target_weights = torch.tensor(args.target_weights).unsqueeze(0)  # shape(1,tasks)
            else:
                target_weights = torch.ones(targets.shape[1]).unsqueeze(0)
            data_weights = torch.tensor(data_weights_batch).unsqueeze(1)  # shape(batch,1)

            constraints_batch = None
            bond_types_batch = None

            if args.loss_function == "bounded_mse":
                lt_target_batch = batch.lt_targets()  # shape(batch, tasks)
                gt_target_batch = batch.gt_targets()  # shape(batch, tasks)
                lt_target_batch = torch.tensor(lt_target_batch)
                gt_target_batch = torch.tensor(gt_target_batch)

        # Run model
        model.zero_grad()
        preds = model(
            mol_batch,
            features_batch,
            atom_descriptors_batch,
            atom_features_batch,
            bond_descriptors_batch,
            bond_features_batch,
            constraints_batch,
            bond_types_batch,
        )

        # Move tensors to correct device
        torch_device = args.device
        if model.is_atom_bond_targets:
            masks = [x.to(torch_device) for x in masks]
            masks = [x.reshape([-1, 1]) for x in masks]
            targets = [x.to(torch_device) for x in targets]
            targets = [x.reshape([-1, 1]) for x in targets]
            target_weights = [x.to(torch_device) for x in target_weights]
            data_weights = [x.to(torch_device) for x in data_weights]
        else:
            masks = masks.to(torch_device)
            targets = targets.to(torch_device)
            target_weights = target_weights.to(torch_device)
            data_weights = data_weights.to(torch_device)
            if args.loss_function == "bounded_mse":
                lt_target_batch = lt_target_batch.to(torch_device)
                gt_target_batch = gt_target_batch.to(torch_device)

        # Calculate losses
        if model.is_atom_bond_targets:
            loss_multi_task = []
            for target, pred, target_weight, data_weight, mask in zip(targets, preds, target_weights, data_weights, masks):
                if args.loss_function == "mcc" and args.dataset_type == "classification":
                    loss = loss_func(pred, target, data_weight, mask) * target_weight.squeeze(0)
                elif args.loss_function == "bounded_mse":
                    raise ValueError(f'Loss function "{args.loss_function}" is not supported with dataset type {args.dataset_type} in atomic/bond properties prediction.')
                elif args.loss_function in ["binary_cross_entropy", "mse", "mve"]:
                    loss = loss_func(pred, target) * target_weight * data_weight * mask
                elif args.loss_function == "evidential":
                    loss = loss_func(pred, target, args.evidential_regularization) * target_weight * data_weight * mask
                elif args.loss_function == "dirichlet" and args.dataset_type == "classification":
                    loss = loss_func(pred, target, args.evidential_regularization) * target_weight * data_weight * mask
                elif args.loss_function == "quantile_interval":
                    quantiles_tensor = torch.tensor(args.quantiles, device=torch_device)
                    loss = loss_func(pred, target, quantiles_tensor) * target_weight * data_weight * mask
                else:
                    raise ValueError(f'Dataset type "{args.dataset_type}" is not supported.')
                loss = loss.sum() / mask.sum()
                loss_multi_task.append(loss)

            loss_sum = [x + y for x, y in zip(loss_sum, loss_multi_task)]
            iter_count += 1

            sum(loss_multi_task).backward()
        else:
            if args.loss_function == "mcc" and args.dataset_type == "classification":
                loss = loss_func(preds, targets, data_weights, masks) * target_weights.squeeze(0)
            elif args.loss_function == "mcc":  # multiclass dataset type
                targets = targets.long()
                target_losses = []
                for target_index in range(preds.size(1)):
                    target_loss = loss_func(preds[:, target_index, :], targets[:, target_index], data_weights, masks[:, target_index]).unsqueeze(0)
                    target_losses.append(target_loss)
                loss = torch.cat(target_losses) * target_weights.squeeze(0)
            elif args.dataset_type == "multiclass":
                targets = targets.long()
                if args.loss_function == "dirichlet":
                    loss = loss_func(preds, targets, args.evidential_regularization) * target_weights * data_weights * masks
                else:
                    target_losses = []
                    for target_index in range(preds.size(1)):
                        target_loss = loss_func(preds[:, target_index, :], targets[:, target_index]).unsqueeze(1)
                        target_losses.append(target_loss)
                    loss = torch.cat(target_losses, dim=1).to(torch_device) * target_weights * data_weights * masks
            elif args.dataset_type == "spectra":
                loss = loss_func(preds, targets, masks) * target_weights * data_weights * masks
            elif args.loss_function == "bounded_mse":
                loss = loss_func(preds, targets, lt_target_batch, gt_target_batch) * target_weights * data_weights * masks
            elif args.loss_function == "evidential":
                loss = loss_func(preds, targets, args.evidential_regularization) * target_weights * data_weights * masks
            elif args.loss_function == "dirichlet":  # classification
                loss = loss_func(preds, targets, args.evidential_regularization) * target_weights * data_weights * masks
            elif args.loss_function == "quantile_interval":
                quantiles_tensor = torch.tensor(args.quantiles, device=torch_device)
                loss = loss_func(preds, targets, quantiles_tensor) * target_weights * data_weights * masks
            else:
                loss = loss_func(preds, targets) * target_weights * data_weights * masks


            if args.loss_function == "mcc":
                loss = loss.mean()
            else:
                loss = loss.sum() / masks.sum()

            loss_sum += loss.item()
            iter_count += 1

            loss.backward()
        if args.grad_clip:
            nn.utils.clip_grad_norm_(model.parameters(), args.grad_clip)
        optimizer.step()

        if isinstance(scheduler, NoamLR):
            scheduler.step()

        n_iter += len(batch)

        # Log and/or add to tensorboard
        if (n_iter // args.batch_size) % args.log_frequency == 0:
            lrs = scheduler.get_lr()
            pnorm = compute_pnorm(model)
            gnorm = compute_gnorm(model)
            if model.is_atom_bond_targets:
                loss_avg = sum(loss_sum) / iter_count
                loss_sum, iter_count = [0]*(len(args.atom_targets) + len(args.bond_targets)), 0
            else:
                loss_avg = loss_sum / iter_count
                loss_sum = iter_count = 0

            lrs_str = ", ".join(f"lr_{i} = {lr:.4e}" for i, lr in enumerate(lrs))
            debug(f"Loss = {loss_avg:.4e}, PNorm = {pnorm:.4f}, GNorm = {gnorm:.4f}, {lrs_str}")

            if writer is not None:
                writer.add_scalar("train_loss", loss_avg, n_iter)
                writer.add_scalar("param_norm", pnorm, n_iter)
                writer.add_scalar("gradient_norm", gnorm, n_iter)
                for i, lr in enumerate(lrs):
                    writer.add_scalar(f"learning_rate_{i}", lr, n_iter)

    return n_iter



from typing import Callable

import torch
import torch.nn as nn
import numpy as np

from chemprop.args import TrainArgs


def get_loss_func(args: TrainArgs) -> Callable:
    """
    Gets the loss function corresponding to a given dataset type.

    :param args: Arguments containing the dataset type ("classification", "regression", or "multiclass").
    :return: A PyTorch loss function.
    """

    # Nested dictionary of the form {dataset_type: {loss_function: loss_function callable}}
    supported_loss_functions = {
        "regression": {
            "mse": nn.MSELoss(reduction="none"),
            "bounded_mse": bounded_mse_loss,
            "mve": normal_mve,
            "evidential": evidential_loss,
            "quantile_interval": quantile_loss,
        },
        "classification": {
            "binary_cross_entropy": nn.BCEWithLogitsLoss(reduction="none"),
            "mcc": mcc_class_loss,
            "dirichlet": dirichlet_class_loss,
        },
        "multiclass": {
            "cross_entropy": nn.CrossEntropyLoss(reduction="none"),
            "mcc": mcc_multiclass_loss,
            "dirichlet": dirichlet_multiclass_loss,
        },
        "spectra": {
            "sid": sid_loss,
            "wasserstein": wasserstein_loss,
        },
    }

    # Error if no loss function supported
    if args.dataset_type not in supported_loss_functions.keys():
        raise ValueError(f'Dataset type "{args.dataset_type}" not supported.')

    # Return loss function if it is represented in the supported_loss_functions dictionary
    loss_function_choices = supported_loss_functions.get(args.dataset_type, dict())
    loss_function = loss_function_choices.get(args.loss_function)

    if loss_function is not None:
        return loss_function

    else:
        raise ValueError(
            f'Loss function "{args.loss_function}" not supported with dataset type {args.dataset_type}. \
            Available options for that dataset type are {loss_function_choices.keys()}.'
        )


def bounded_mse_loss(
    predictions: torch.tensor,
    targets: torch.tensor,
    less_than_target: torch.tensor,
    greater_than_target: torch.tensor,
) -> torch.tensor:
    """
    Loss function for use with regression when some targets are presented as inequalities.

    :param predictions: Model predictions with shape(batch_size, tasks).
    :param targets: Target values with shape(batch_size, tasks).
    :param less_than_target: A tensor with boolean values indicating whether the target is a less-than inequality.
    :param greater_than_target: A tensor with boolean values indicating whether the target is a greater-than inequality.
    :return: A tensor containing loss values of shape(batch_size, tasks).
    """
    predictions = torch.where(torch.logical_and(predictions < targets, less_than_target), targets, predictions)

    predictions = torch.where(
        torch.logical_and(predictions > targets, greater_than_target),
        targets,
        predictions,
    )

    return nn.functional.mse_loss(predictions, targets, reduction="none")


def mcc_class_loss(
    predictions: torch.tensor,
    targets: torch.tensor,
    data_weights: torch.tensor,
    mask: torch.tensor,
) -> torch.tensor:
    """
    A classification loss using a soft version of the Matthews Correlation Coefficient.

    :param predictions: Model predictions with shape(batch_size, tasks).
    :param targets: Target values with shape(batch_size, tasks).
    :param data_weights: A tensor with float values indicating how heavily to weight each datapoint in training with shape(batch_size, 1)
    :param mask: A tensor with boolean values indicating whether the loss for this prediction is considered in the gradient descent with shape(batch_size, tasks).
    :return: A tensor containing loss values of shape(tasks).
    """
    # shape(batch, tasks)
    # (TP*TN-FP*FN)/sqrt((TP+FP)*(TP+FN)*(TN+FP)*(TN+FN))
    TP = torch.sum(targets * predictions * data_weights * mask, axis=0)
    FP = torch.sum((1 - targets) * predictions * data_weights * mask, axis=0)
    FN = torch.sum(targets * (1 - predictions) * data_weights * mask, axis=0)
    TN = torch.sum((1 - targets) * (1 - predictions) * data_weights * mask, axis=0)
    loss = 1 - ((TP * TN - FP * FN) / torch.sqrt((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN)))
    return loss


def mcc_multiclass_loss(
    predictions: torch.tensor,
    targets: torch.tensor,
    data_weights: torch.tensor,
    mask: torch.tensor,
) -> torch.tensor:
    """
    A multiclass loss using a soft version of the Matthews Correlation Coefficient. Multiclass definition follows the version in sklearn documentation (https://scikit-learn.org/stable/modules/model_evaluation.html#matthews-correlation-coefficient).

    :param predictions: Model predictions with shape(batch_size, classes).
    :param targets: Target values with shape(batch_size).
    :param data_weights: A tensor with float values indicating how heavily to weight each datapoint in training with shape(batch_size, 1)
    :param mask: A tensor with boolean values indicating whether the loss for this prediction is considered in the gradient descent with shape(batch_size).
    :return: A tensor value for the loss.
    """
    torch_device = predictions.device
    mask = mask.unsqueeze(1)

    bin_targets = torch.zeros_like(predictions, device=torch_device)
    bin_targets[torch.arange(predictions.shape[0]), targets] = 1

    pred_classes = predictions.argmax(dim=1)
    bin_preds = torch.zeros_like(predictions, device=torch_device)
    bin_preds[torch.arange(predictions.shape[0]), pred_classes] = 1

    masked_data_weights = data_weights * mask

    t_sum = torch.sum(bin_targets * masked_data_weights, axis=0)  # number of times each class truly occurred
    p_sum = torch.sum(bin_preds * masked_data_weights, axis=0)  # number of times each class was predicted

    n_correct = torch.sum(bin_preds * bin_targets * masked_data_weights)  # total number of samples correctly predicted
    n_samples = torch.sum(predictions * masked_data_weights)  # total number of samples

    cov_ytyp = n_correct * n_samples - torch.dot(p_sum, t_sum)
    cov_ypyp = n_samples**2 - torch.dot(p_sum, p_sum)
    cov_ytyt = n_samples**2 - torch.dot(t_sum, t_sum)

    if cov_ypyp * cov_ytyt == 0:
        loss = torch.tensor(1.0, device=torch_device)
    else:
        mcc = cov_ytyp / torch.sqrt(cov_ytyt * cov_ypyp)
        loss = 1 - mcc

    return loss


def sid_loss(
    model_spectra: torch.tensor,
    target_spectra: torch.tensor,
    mask: torch.tensor,
    threshold: float = None,
) -> torch.tensor:
    """
    Loss function for use with spectra data type.

    :param model_spectra: The predicted spectra output from a model with shape (batch_size,spectrum_length).
    :param target_spectra: The target spectra with shape (batch_size,spectrum_length). Values must be normalized so that each spectrum sums to 1.
    :param mask: Tensor with boolean indications of where the spectrum output should not be excluded with shape (batch_size,spectrum_length).
    :param threshold: Loss function requires that values are positive and nonzero. Values below the threshold will be replaced with the threshold value.
    :return: A tensor containing loss values for the batch with shape (batch_size,spectrum_length).
    """
    # Move new tensors to torch device
    torch_device = model_spectra.device

    # Normalize the model spectra before comparison
    zero_sub = torch.zeros_like(model_spectra, device=torch_device)
    one_sub = torch.ones_like(model_spectra, device=torch_device)
    if threshold is not None:
        threshold_sub = torch.full(model_spectra.shape, threshold, device=torch_device)
        model_spectra = torch.where(model_spectra < threshold, threshold_sub, model_spectra)
    model_spectra = torch.where(mask, model_spectra, zero_sub)
    sum_model_spectra = torch.sum(model_spectra, axis=1, keepdim=True)
    model_spectra = torch.div(model_spectra, sum_model_spectra)

    # Calculate loss value
    target_spectra = torch.where(mask, target_spectra, one_sub)
    model_spectra = torch.where(mask, model_spectra, one_sub)  # losses in excluded regions will be zero because log(1/1) = 0.
    loss = torch.mul(torch.log(torch.div(model_spectra, target_spectra)), model_spectra) + torch.mul(
        torch.log(torch.div(target_spectra, model_spectra)), target_spectra
    )

    return loss


def wasserstein_loss(
    model_spectra: torch.tensor,
    target_spectra: torch.tensor,
    mask: torch.tensor,
    threshold: float = None,
) -> torch.tensor:
    """
    Loss function for use with spectra data type. This loss assumes that values are evenly spaced.

    :param model_spectra: The predicted spectra output from a model with shape (batch_size,spectrum_length).
    :param target_spectra: The target spectra with shape (batch_size,spectrum_length). Values must be normalized so that each spectrum sums to 1.
    :param mask: Tensor with boolian indications of where the spectrum output should not be excluded with shape (batch_size,spectrum_length).
    :param threshold: Loss function requires that values are positive and nonzero. Values below the threshold will be replaced with the threshold value.
    :return: A tensor containing loss values for the batch with shape (batch_size,spectrum_length).
    """
    # Move new tensors to torch device
    torch_device = model_spectra.device

    # Normalize the model spectra before comparison
    zero_sub = torch.zeros_like(model_spectra, device=torch_device)
    if threshold is not None:
        threshold_sub = torch.full(model_spectra.shape, threshold, device=torch_device)
        model_spectra = torch.where(model_spectra < threshold, threshold_sub, model_spectra)
    model_spectra = torch.where(mask, model_spectra, zero_sub)
    sum_model_spectra = torch.sum(model_spectra, axis=1, keepdim=True)
    model_spectra = torch.div(model_spectra, sum_model_spectra)

    # Calculate loss value
    target_cum = torch.cumsum(target_spectra, axis=1)
    model_cum = torch.cumsum(model_spectra, axis=1)
    loss = torch.abs(target_cum - model_cum)

    return loss


def normal_mve(pred_values, targets):
    """
    Use the negative log likelihood function of a normal distribution as a loss function used for making
    simultaneous predictions of the mean and error distribution variance simultaneously.

    :param pred_values: Combined predictions of means and variances of shape(data, tasks*2).
                        Means are first in dimension 1, followed by variances.
    :return: A tensor loss value.
    """
    # Unpack combined prediction values
    pred_means, pred_var = torch.split(pred_values, pred_values.shape[1] // 2, dim=1)

    return torch.log(2 * np.pi * pred_var) / 2 + (pred_means - targets) ** 2 / (2 * pred_var)


# evidential classification
def dirichlet_class_loss(alphas, target_labels, lam=0):
    """
    Use Evidential Learning Dirichlet loss from Sensoy et al in classification datasets.
    :param alphas: Predicted parameters for Dirichlet in shape(datapoints, tasks*2).
                   Negative class first then positive class in dimension 1.
    :param target_labels: Digital labels to predict in shape(datapoints, tasks).
    :lambda: coefficient to weight KL term

    :return: Loss
    """
    torch_device = alphas.device
    num_tasks = target_labels.shape[1]
    num_classes = 2
    alphas = torch.reshape(alphas, (alphas.shape[0], num_tasks, num_classes))

    y_one_hot = torch.eye(num_classes, device=torch_device)[target_labels.long()]

    return dirichlet_common_loss(alphas=alphas, y_one_hot=y_one_hot, lam=lam)


def dirichlet_multiclass_loss(alphas, target_labels, lam=0):
    """
    Use Evidential Learning Dirichlet loss from Sensoy et al for multiclass datasets.
    :param alphas: Predicted parameters for Dirichlet in shape(datapoints, task, classes).
    :param target_labels: Digital labels to predict in shape(datapoints, tasks).
    :lambda: coefficient to weight KL term

    :return: Loss
    """
    torch_device = alphas.device
    num_classes = alphas.shape[2]

    y_one_hot = torch.eye(num_classes, device=torch_device)[target_labels.long()]

    return dirichlet_common_loss(alphas=alphas, y_one_hot=y_one_hot, lam=lam)


def dirichlet_common_loss(alphas, y_one_hot, lam=0):
    """
    Use Evidential Learning Dirichlet loss from Sensoy et al. This function follows
    after the classification and multiclass specific functions that reshape the
    alpha inputs and create one-hot targets.

    :param alphas: Predicted parameters for Dirichlet in shape(datapoints, task, classes).
    :param y_one_hot: Digital labels to predict in shape(datapoints, tasks, classes).
    :lambda: coefficient to weight KL term

    :return: Loss
    """
    # SOS term
    S = torch.sum(alphas, dim=-1, keepdim=True)
    p = alphas / S
    A = torch.sum((y_one_hot - p) ** 2, dim=-1, keepdim=True)
    B = torch.sum((p * (1 - p)) / (S + 1), dim=-1, keepdim=True)
    SOS = A + B

    alpha_hat = y_one_hot + (1 - y_one_hot) * alphas

    beta = torch.ones_like(alpha_hat)
    S_alpha = torch.sum(alpha_hat, dim=-1, keepdim=True)
    S_beta = torch.sum(beta, dim=-1, keepdim=True)

    ln_alpha = torch.lgamma(S_alpha) - torch.sum(torch.lgamma(alpha_hat), dim=-1, keepdim=True)
    ln_beta = torch.sum(torch.lgamma(beta), dim=-1, keepdim=True) - torch.lgamma(S_beta)

    # digamma terms
    dg_alpha = torch.digamma(alpha_hat)
    dg_S_alpha = torch.digamma(S_alpha)

    # KL
    KL = ln_alpha + ln_beta + torch.sum((alpha_hat - beta) * (dg_alpha - dg_S_alpha), dim=-1, keepdim=True)

    KL = lam * KL

    # loss = torch.mean(SOS + KL)
    loss = SOS + KL
    loss = torch.mean(loss, dim=-1)
    return loss


# updated evidential regression loss (evidential_loss_new from Amini repo)
def evidential_loss(pred_values, targets, lam: float = 0, epsilon: float = 1e-8, v_min: float = 1e-5):
    """
    Use Deep Evidential Regression negative log likelihood loss + evidential
        regularizer

    :param pred_values: Combined prediction values for mu, v, alpha, and beta parameters in shape(data, tasks*4).
                        Order in dimension 1 is mu, v, alpha, beta.
    :mu: pred mean parameter for NIG
    :v: pred lam parameter for NIG
    :alpha: predicted parameter for NIG
    :beta: Predicted parmaeter for NIG
    :param targets: Outputs to predict
    :param lam: regularization coefficient
    :param v_min: clamp any v below this value to prevent Inf from division

    :return: Loss
    """
    # Unpack combined prediction values
    mu, v, alpha, beta = torch.split(pred_values, pred_values.shape[1] // 4, dim=1)

    # Calculate NLL loss
    v = torch.clamp(v, v_min)
    twoBlambda = 2 * beta * (1 + v)
    nll = (
        0.5 * torch.log(np.pi / v)
        - alpha * torch.log(twoBlambda)
        + (alpha + 0.5) * torch.log(v * (targets - mu) ** 2 + twoBlambda)
        + torch.lgamma(alpha)
        - torch.lgamma(alpha + 0.5)
    )

    L_NLL = nll  # torch.mean(nll, dim=-1)

    # Calculate regularizer based on absolute error of prediction
    error = torch.abs((targets - mu))
    reg = error * (2 * v + alpha)
    L_REG = reg  # torch.mean(reg, dim=-1)

    # Loss = L_NLL + L_REG
    # TODO If we want to optimize the dual- of the objective use the line below:
    loss = L_NLL + lam * (L_REG - epsilon)

    return loss


def quantile_loss(pred_values: torch.Tensor, targets: torch.Tensor, quantiles: torch.Tensor):
    """
    Batched pinball loss at desired quantiles.
    """

    error = pred_values - targets

    return torch.max((1 - quantiles) * error, -quantiles * error)



"""Contains utility functions for the Flask web app."""

import os
import shutil

from flask import Flask


def set_root_folder(app: Flask, root_folder: str = None, create_folders: bool = True) -> None:
    """
    Sets the root folder for the config along with subfolders like the data and checkpoint folders.

    :param app: Flask app.
    :param root_folder: Path to the root folder. If None, the current root folders is unchanged.
    :param create_folders: Whether to create the root folder and subfolders.
    """
    # Set root folder and subfolders
    if root_folder is not None:
        app.config['ROOT_FOLDER'] = root_folder
        app.config['DATA_FOLDER'] = os.path.join(app.config['ROOT_FOLDER'], 'app/web_data')
        app.config['CHECKPOINT_FOLDER'] = os.path.join(app.config['ROOT_FOLDER'], 'app/web_checkpoints')
        app.config['TEMP_FOLDER'] = os.path.join(app.config['ROOT_FOLDER'], 'app/temp')
        app.config['DB_PATH'] = os.path.join(app.config['ROOT_FOLDER'], app.config['DB_FILENAME'])

    # Create folders
    if create_folders:
        if not os.access(os.path.dirname(app.config['ROOT_FOLDER']), os.W_OK):
            raise ValueError(f'You do not have write permissions on the root_folder: {app.config["ROOT_FOLDER"]}\n'
                             f'Please specify a different root_folder while starting the web app.')

        for folder_name in ['ROOT_FOLDER', 'DATA_FOLDER', 'CHECKPOINT_FOLDER', 'TEMP_FOLDER']:
            os.makedirs(app.config[folder_name], exist_ok=True)


def clear_temp_folder(app: Flask) -> None:
    """Clears the temporary folder."""
    shutil.rmtree(app.config['TEMP_FOLDER'])
    os.makedirs(app.config['TEMP_FOLDER'], exist_ok=True)



from typing import List, Union, Tuple

import numpy as np
from rdkit import Chem
import torch
import torch.nn as nn

from .mpn import MPN
from .ffn import build_ffn, MultiReadout
from chemprop.args import TrainArgs
from chemprop.features import BatchMolGraph
from chemprop.nn_utils import initialize_weights


class MoleculeModel(nn.Module):
    """A :class:`MoleculeModel` is a model which contains a message passing network following by feed-forward layers."""

    def __init__(self, args: TrainArgs):
        """
        :param args: A :class:`~chemprop.args.TrainArgs` object containing model arguments.
        """
        super(MoleculeModel, self).__init__()

        self.classification = args.dataset_type == "classification"
        self.multiclass = args.dataset_type == "multiclass"
        self.loss_function = args.loss_function

        if hasattr(args, "train_class_sizes"):
            self.train_class_sizes = args.train_class_sizes
        else:
            self.train_class_sizes = None

        # when using cross entropy losses, no sigmoid or softmax during training. But they are needed for mcc loss.
        if self.classification or self.multiclass:
            self.no_training_normalization = args.loss_function in [
                "cross_entropy",
                "binary_cross_entropy",
            ]

        self.is_atom_bond_targets = args.is_atom_bond_targets

        if self.is_atom_bond_targets:
            self.atom_targets, self.bond_targets = args.atom_targets, args.bond_targets
            self.atom_constraints, self.bond_constraints = (
                args.atom_constraints,
                args.bond_constraints,
            )
            self.adding_bond_types = args.adding_bond_types

        self.relative_output_size = 1
        if self.multiclass:
            self.relative_output_size *= args.multiclass_num_classes
        if self.loss_function == "mve":
            self.relative_output_size *= 2  # return means and variances
        if self.loss_function == "dirichlet" and self.classification:
            self.relative_output_size *= (
                2  # return dirichlet parameters for positive and negative class
            )
        if self.loss_function == "evidential":
            self.relative_output_size *= (
                4  # return four evidential parameters: gamma, lambda, alpha, beta
            )

        if self.classification:
            self.sigmoid = nn.Sigmoid()

        if self.multiclass:
            self.multiclass_softmax = nn.Softmax(dim=2)

        if self.loss_function in ["mve", "evidential", "dirichlet"]:
            self.softplus = nn.Softplus()

        self.create_encoder(args)
        self.create_ffn(args)

        initialize_weights(self)

    def create_encoder(self, args: TrainArgs) -> None:
        """
        Creates the message passing encoder for the model.

        :param args: A :class:`~chemprop.args.TrainArgs` object containing model arguments.
        """
        self.encoder = MPN(args)

        if args.checkpoint_frzn is not None:
            if args.freeze_first_only:  # Freeze only the first encoder
                for param in list(self.encoder.encoder.children())[0].parameters():
                    param.requires_grad = False
            else:  # Freeze all encoders
                for param in self.encoder.parameters():
                    param.requires_grad = False

    def create_ffn(self, args: TrainArgs) -> None:
        """
        Creates the feed-forward layers for the model.

        :param args: A :class:`~chemprop.args.TrainArgs` object containing model arguments.
        """
        self.multiclass = args.dataset_type == "multiclass"
        if self.multiclass:
            self.num_classes = args.multiclass_num_classes
        if args.features_only:
            first_linear_dim = args.features_size
        else:
            if args.reaction_solvent:
                first_linear_dim = args.hidden_size + args.hidden_size_solvent
            else:
                first_linear_dim = args.hidden_size * args.number_of_molecules
            if args.use_input_features:
                first_linear_dim += args.features_size

        if args.atom_descriptors == "descriptor":
            atom_first_linear_dim = first_linear_dim + args.atom_descriptors_size
        else:
            atom_first_linear_dim = first_linear_dim

        if args.bond_descriptors == "descriptor":
            bond_first_linear_dim = first_linear_dim + args.bond_descriptors_size
        else:
            bond_first_linear_dim = first_linear_dim

        # Create FFN layers
        if self.is_atom_bond_targets:
            self.readout = MultiReadout(
                atom_features_size=atom_first_linear_dim,
                bond_features_size=bond_first_linear_dim,
                atom_hidden_size=args.ffn_hidden_size + args.atom_descriptors_size,
                bond_hidden_size=args.ffn_hidden_size + args.bond_descriptors_size,
                num_layers=args.ffn_num_layers,
                output_size=self.relative_output_size,
                dropout=args.dropout,
                activation=args.activation,
                atom_constraints=args.atom_constraints,
                bond_constraints=args.bond_constraints,
                shared_ffn=args.shared_atom_bond_ffn,
                weights_ffn_num_layers=args.weights_ffn_num_layers,
            )
        else:
            self.readout = build_ffn(
                first_linear_dim=atom_first_linear_dim,
                hidden_size=args.ffn_hidden_size + args.atom_descriptors_size,
                num_layers=args.ffn_num_layers,
                output_size=self.relative_output_size * args.num_tasks,
                dropout=args.dropout,
                activation=args.activation,
                dataset_type=args.dataset_type,
                spectra_activation=args.spectra_activation,
            )

        if args.checkpoint_frzn is not None:
            if args.frzn_ffn_layers > 0:
                if self.is_atom_bond_targets:
                    if args.shared_atom_bond_ffn:
                        for param in list(self.readout.atom_ffn_base.parameters())[
                            0 : 2 * args.frzn_ffn_layers
                        ]:
                            param.requires_grad = False
                        for param in list(self.readout.bond_ffn_base.parameters())[
                            0 : 2 * args.frzn_ffn_layers
                        ]:
                            param.requires_grad = False
                    else:
                        for ffn in self.readout.ffn_list:
                            if ffn.constraint:
                                for param in list(ffn.ffn.parameters())[
                                    0 : 2 * args.frzn_ffn_layers
                                ]:
                                    param.requires_grad = False
                            else:
                                for param in list(ffn.ffn_readout.parameters())[
                                    0 : 2 * args.frzn_ffn_layers
                                ]:
                                    param.requires_grad = False
                else:
                    for param in list(self.readout.parameters())[
                        0 : 2 * args.frzn_ffn_layers
                    ]:  # Freeze weights and bias for given number of layers
                        param.requires_grad = False

    def fingerprint(
        self,
        batch: Union[
            List[List[str]],
            List[List[Chem.Mol]],
            List[List[Tuple[Chem.Mol, Chem.Mol]]],
            List[BatchMolGraph],
        ],
        features_batch: List[np.ndarray] = None,
        atom_descriptors_batch: List[np.ndarray] = None,
        atom_features_batch: List[np.ndarray] = None,
        bond_descriptors_batch: List[np.ndarray] = None,
        bond_features_batch: List[np.ndarray] = None,
        fingerprint_type: str = "MPN",
    ) -> torch.Tensor:
        """
        Encodes the latent representations of the input molecules from intermediate stages of the model.

        :param batch: A list of list of SMILES, a list of list of RDKit molecules, or a
                      list of :class:`~chemprop.features.featurization.BatchMolGraph`.
                      The outer list or BatchMolGraph is of length :code:`num_molecules` (number of datapoints in batch),
                      the inner list is of length :code:`number_of_molecules` (number of molecules per datapoint).
        :param features_batch: A list of numpy arrays containing additional features.
        :param atom_descriptors_batch: A list of numpy arrays containing additional atom descriptors.
        :param atom_features_batch: A list of numpy arrays containing additional atom features.
        :param bond_descriptors_batch: A list of numpy arrays containing additional bond descriptors.
        :param bond_features_batch: A list of numpy arrays containing additional bond features.
        :param fingerprint_type: The choice of which type of latent representation to return as the molecular fingerprint. Currently
                                 supported MPN for the output of the MPNN portion of the model or last_FFN for the input to the final readout layer.
        :return: The latent fingerprint vectors.
        """
        if fingerprint_type == "MPN":
            return self.encoder(
                batch,
                features_batch,
                atom_descriptors_batch,
                atom_features_batch,
                bond_descriptors_batch,
                bond_features_batch,
            )
        elif fingerprint_type == "last_FFN":
            return self.readout[:-1](
                self.encoder(
                    batch,
                    features_batch,
                    atom_descriptors_batch,
                    atom_features_batch,
                    bond_descriptors_batch,
                    bond_features_batch,
                )
            )
        else:
            raise ValueError(f"Unsupported fingerprint type {fingerprint_type}.")

    def forward(
        self,
        batch: Union[
            List[List[str]],
            List[List[Chem.Mol]],
            List[List[Tuple[Chem.Mol, Chem.Mol]]],
            List[BatchMolGraph],
        ],
        features_batch: List[np.ndarray] = None,
        atom_descriptors_batch: List[np.ndarray] = None,
        atom_features_batch: List[np.ndarray] = None,
        bond_descriptors_batch: List[np.ndarray] = None,
        bond_features_batch: List[np.ndarray] = None,
        constraints_batch: List[torch.Tensor] = None,
        bond_types_batch: List[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Runs the :class:`MoleculeModel` on input.

        :param batch: A list of list of SMILES, a list of list of RDKit molecules, or a
                      list of :class:`~chemprop.features.featurization.BatchMolGraph`.
                      The outer list or BatchMolGraph is of length :code:`num_molecules` (number of datapoints in batch),
                      the inner list is of length :code:`number_of_molecules` (number of molecules per datapoint).
        :param features_batch: A list of numpy arrays containing additional features.
        :param atom_descriptors_batch: A list of numpy arrays containing additional atom descriptors.
        :param atom_features_batch: A list of numpy arrays containing additional atom features.
        :param bond_descriptors_batch: A list of numpy arrays containing additional bond descriptors.
        :param bond_features_batch: A list of numpy arrays containing additional bond features.
        :param constraints_batch: A list of PyTorch tensors which applies constraint on atomic/bond properties.
        :param bond_types_batch: A list of PyTorch tensors storing bond types of each bond determined by RDKit molecules.
        :return: The output of the :class:`MoleculeModel`, containing a list of property predictions.
        """
        if self.is_atom_bond_targets:
            encodings = self.encoder(
                batch,
                features_batch,
                atom_descriptors_batch,
                atom_features_batch,
                bond_descriptors_batch,
                bond_features_batch,
            )
            output = self.readout(encodings, constraints_batch, bond_types_batch)
        else:
            encodings = self.encoder(
                batch,
                features_batch,
                atom_descriptors_batch,
                atom_features_batch,
                bond_descriptors_batch,
                bond_features_batch,
            )
            output = self.readout(encodings)

        # Don't apply sigmoid during training when using BCEWithLogitsLoss
        if (
            self.classification
            and not (self.training and self.no_training_normalization)
            and self.loss_function != "dirichlet"
        ):
            if self.is_atom_bond_targets:
                output = [self.sigmoid(x) for x in output]
            else:
                output = self.sigmoid(output)
        if self.multiclass:
            output = output.reshape(
                (output.shape[0], -1, self.num_classes)
            )  # batch size x num targets x num classes per target
            if (
                not (self.training and self.no_training_normalization)
                and self.loss_function != "dirichlet"
            ):
                output = self.multiclass_softmax(
                    output
                )  # to get probabilities during evaluation, but not during training when using CrossEntropyLoss

        # Modify multi-input loss functions
        if self.loss_function == "mve":
            if self.is_atom_bond_targets:
                outputs = []
                for x in output:
                    means, variances = torch.split(x, x.shape[1] // 2, dim=1)
                    variances = self.softplus(variances)
                    outputs.append(torch.cat([means, variances], axis=1))
                return outputs
            else:
                means, variances = torch.split(output, output.shape[1] // 2, dim=1)
                variances = self.softplus(variances)
                output = torch.cat([means, variances], axis=1)
        if self.loss_function == "evidential":
            if self.is_atom_bond_targets:
                outputs = []
                for x in output:
                    means, lambdas, alphas, betas = torch.split(
                        x, x.shape[1] // 4, dim=1
                    )
                    lambdas = self.softplus(lambdas)  # + min_val
                    alphas = (
                        self.softplus(alphas) + 1
                    )  # + min_val # add 1 for numerical contraints of Gamma function
                    betas = self.softplus(betas)  # + min_val
                    outputs.append(torch.cat([means, lambdas, alphas, betas], dim=1))
                return outputs
            else:
                means, lambdas, alphas, betas = torch.split(
                    output, output.shape[1] // 4, dim=1
                )
                lambdas = self.softplus(lambdas)  # + min_val
                alphas = (
                    self.softplus(alphas) + 1
                )  # + min_val # add 1 for numerical contraints of Gamma function
                betas = self.softplus(betas)  # + min_val
                output = torch.cat([means, lambdas, alphas, betas], dim=1)
        if self.loss_function == "dirichlet":
            if self.is_atom_bond_targets:
                outputs = []
                for x in output:
                    outputs.append(nn.functional.softplus(x) + 1)
                return outputs
            else:
                output = nn.functional.softplus(output) + 1

        return output



"""Runs the web interface version of chemprop, allowing for training and predicting in a web browser."""
import os

from flask import Flask

from chemprop.web.utils import set_root_folder


app = Flask(__name__)
app.config.from_object('chemprop.web.config')
set_root_folder(
    app=app,
    root_folder=os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
    create_folders=False
)

from chemprop.web.app import views



from setuptools import find_packages, setup

__version__ = "1.7.0"

# Load README
with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="chemprop",
    author="The Chemprop Development Team (see LICENSE.txt)",
    author_email="chemprop@mit.edu",
    description="Molecular Property Prediction with Message Passing Neural Networks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/chemprop/chemprop",
    download_url=f"https://github.com/chemprop/chemprop/v_{__version__}.tar.gz",
    project_urls={
        "Documentation": "https://chemprop.readthedocs.io/en/latest/",
        "Source": "https://github.com/chemprop/chemprop",
        "PyPi": "https://pypi.org/project/chemprop/",
    },
    license="MIT",
    packages=find_packages(),
    package_data={"chemprop": ["py.typed"]},
    entry_points={
        "console_scripts": [
            "chemprop_train=chemprop.train:chemprop_train",
            "chemprop_predict=chemprop.train:chemprop_predict",
            "chemprop_fingerprint=chemprop.train:chemprop_fingerprint",
            "chemprop_hyperopt=chemprop.hyperparameter_optimization:chemprop_hyperopt",
            "chemprop_interpret=chemprop.interpret:chemprop_interpret",
            "chemprop_web=chemprop.web.run:chemprop_web",
            "sklearn_train=chemprop.sklearn_train:sklearn_train",
            "sklearn_predict=chemprop.sklearn_predict:sklearn_predict",
        ]
    },
    install_requires=[
        "flask>=1.1.2,<=2.1.3",
        "Werkzeug<3",
        "hyperopt>=0.2.3",
        "matplotlib>=3.1.3",
        "numpy>=1.18.1",
        "pandas>=1.0.3",
        "pandas-flavor>=0.2.0",
        "scikit-learn>=0.22.2.post1",
        "sphinx>=3.1.2",
        "sphinx-rtd-theme>=2.0.0",
        "tensorboardX>=2.0",
        "torch>=1.4.0",
        "tqdm>=4.45.0",
        "typed-argument-parser>=1.6.1",
        "rdkit>=2020.03.1.0",
        "scipy<1.11 ; python_version=='3.7'",
        "descriptastorus<2.6.1 ; python_version=='3.7'",
        "scipy>=1.9 ; python_version=='3.8'",
        "descriptastorus>=2.6.1 ; python_version=='3.8'",
    ],
    extras_require={"test": ["pytest>=6.2.2", "parameterized>=0.8.1"]},
    python_requires=">=3.7,<3.9",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords=[
        "chemistry",
        "machine learning",
        "property prediction",
        "message passing neural network",
        "graph neural network",
    ],
)
